# coding=utf-8
import dbm
import logging
from datetime import datetime
from enum import Enum
from typing import List, TypeVar, Callable, Optional, Any

from googleapiclient.errors import HttpError
from pkg_resources import resource_filename
import struct
from urllib import request, parse
from time import sleep
from urllib.error import HTTPError

import io
import os
import httplib2

from oauth2client.tools import argparser
from oauth2client import client
from oauth2client import tools
from oauth2client import service_account

from oauth2client.file import Storage
from apiclient import discovery

from googleapiclient.http import MediaIoBaseDownload

import mammut
from mammut.common.util import get_datetime_from_rfc3339
from mammut.common import GOOGLE_API_AUTH_LOCAL_WEBSERVER
from mammut.common import GOOGLE_API_AUTH_LOCAL_CREDENTIAL


logger = logging.getLogger()


_FuncT = TypeVar("_FuncT")


class GoogleAPIOpType(Enum):
    READ = "Read"
    WRITE = "Write"
    NGRAM = "Ngram"


class GoogleAPI:
    SCOPES = [
        "https://www.googleapis.com/auth/spreadsheets",
        "https://www.googleapis.com/auth/drive",
        "https://www.googleapis.com/auth/presentations.readonly",
    ]
    CLIENT_SECRET_FILE = "GoogleSheetsMammutEdilmoCredentials.json"
    SERVICE_ACCOUNT_SECRET_FILE = "mammut-devops-cicd.json"
    APPLICATION_NAME = "Mammut"
    slides_metadata_re_string = "(sheet-id|Sheet-id|SHEET-ID)[ ]*->[ ]*([a-zA-Z0-9-_]+)"
    # List of last operations of the same type executions.
    # The list will never have more than the limit.
    READ_EXECUTIONS_TIME_LIST: List[datetime] = [
        mammut.get_timezone_aware_datetime_now()
    ]
    WRITE_EXECUTIONS_TIME_LIST: List[datetime] = [
        mammut.get_timezone_aware_datetime_now()
    ]
    NGRAM_EXECUTIONS_TIME_LIST: List[datetime] = [
        mammut.get_timezone_aware_datetime_now()
    ]
    EXECUTIONS_TIME_LIST = {
        GoogleAPIOpType.READ: READ_EXECUTIONS_TIME_LIST,
        GoogleAPIOpType.WRITE: WRITE_EXECUTIONS_TIME_LIST,
        GoogleAPIOpType.NGRAM: NGRAM_EXECUTIONS_TIME_LIST,
    }
    # Max amount of operations of the same type per time limit
    READ_EXECUTION_LIMIT: int = 100
    WRITE_EXECUTION_LIMIT: int = 100
    NGRAM_EXECUTION_LIMIT: int = 10
    EXECUTION_LIMIT = {
        GoogleAPIOpType.READ: READ_EXECUTION_LIMIT,
        GoogleAPIOpType.WRITE: WRITE_EXECUTION_LIMIT,
        GoogleAPIOpType.NGRAM: NGRAM_EXECUTION_LIMIT,
    }
    # Time spam for max amount of continues operations of the same type in seconds
    READ_EXECUTION_TIME_LIMIT: int = 100
    WRITE_EXECUTION_TIME_LIMIT: int = 100
    NGRAM_EXECUTION_TIME_LIMIT: int = 1
    EXECUTION_TIME_LIMIT = {
        GoogleAPIOpType.READ: READ_EXECUTION_TIME_LIMIT,
        GoogleAPIOpType.WRITE: WRITE_EXECUTION_TIME_LIMIT,
        GoogleAPIOpType.NGRAM: NGRAM_EXECUTION_TIME_LIMIT,
    }
    # Max amount of retries
    MAX_READ_RETRIES: int = 100
    MAX_WRITE_RETRIES: int = 100
    MAX_NGRAM_RETRIES: int = 100
    MAX_RETRIES = {
        GoogleAPIOpType.READ: MAX_READ_RETRIES,
        GoogleAPIOpType.WRITE: MAX_WRITE_RETRIES,
        GoogleAPIOpType.NGRAM: NGRAM_EXECUTION_TIME_LIMIT,
    }
    # Min delay in seconds
    MIN_API_OP_DELAY: float = 1.50
    # Ratio of EXECUTION_TIME_LIMIT to wait after a failure
    EXECUTION_TIME_LIMIT_PROPORTION = 0.3
    CREDENTIALS = None
    SHEETS_SERVICE = None
    DRIVE_SERVICE = None
    SLIDES_SERVICE = None
    IS_AUTHENTICATED = False

    @classmethod
    def ensure_credentials(cls, func: Callable[[_FuncT], _FuncT]):
        def wrapped(*args, **kwargs):
            cls.get_credentials()
            return func(*args, **kwargs)

        return wrapped

    @classmethod
    def ensure_credentials_and_read_limits(cls, func: Callable[[_FuncT], _FuncT]):
        return cls._get_wrapped(func, GoogleAPIOpType.READ)

    @classmethod
    def ensure_credentials_and_write_limits(cls, func: Callable[[_FuncT], _FuncT]):
        return cls._get_wrapped(func, GoogleAPIOpType.WRITE)

    @classmethod
    def ensure_credentials_and_ngram_limits(cls, func: Callable[[_FuncT], _FuncT]):
        return cls._get_wrapped(func, GoogleAPIOpType.NGRAM, 0.0)

    @classmethod
    def _get_wrapped(
        cls,
        func: Callable[[_FuncT], _FuncT],
        op_type: GoogleAPIOpType,
        suppress_exception_value: Optional[Any] = None,
    ):
        def is_exceeded_quote_error(http_error: Exception) -> bool:
            res = False
            if isinstance(http_error, HttpError):
                res = str(http_error.resp.status) == "429"
            elif isinstance(http_error, HTTPError):
                res = http_error.code == 429
            if not res:
                http_error_str = str(http_error).lower()
                res = "quota exceeded" in http_error_str
            return res

        def wrapped(*args, **kwargs):
            cls.get_credentials()
            retries = 0
            result = None
            while result is None:
                cls._check_exec_limit(op_type)

                try:
                    result = func(*args, **kwargs)
                except (HttpError, HTTPError) as http_error:
                    result = None
                    if is_exceeded_quote_error(http_error):
                        retries += 1
                        if retries > cls.MAX_RETRIES[op_type]:
                            mess = "Max amount of google read retries reached"
                            logger.error(mess)
                            raise RuntimeError(mess)
                        else:
                            # If we got an exception is because we didn't wait long enough
                            sleep(
                                cls.EXECUTION_TIME_LIMIT[op_type]
                                * cls.EXECUTION_TIME_LIMIT_PROPORTION
                                * retries
                            )
                            logger.debug(
                                f"Retrying google read operation: {retries}/{cls.MAX_RETRIES[op_type]}"
                            )
                    else:
                        logger.error(
                            f"Unexpected Http Error in read operation: {func.__name__} - Args: {args} - {kwargs}",
                            exc_info=True,
                        )
                        if suppress_exception_value is not None:
                            result = suppress_exception_value
                        else:
                            raise http_error
                except Exception as ex:
                    logger.error(
                        "Unexpected Error in read operation: {func.__name__} - Args: {args} - {kwargs}",
                        exc_info=True,
                    )
                    if suppress_exception_value is not None:
                        result = suppress_exception_value
                    else:
                        raise ex
                finally:
                    current_time: datetime = mammut.get_timezone_aware_datetime_now()
                    cls.EXECUTIONS_TIME_LIST[op_type].append(current_time)
            return result

        return wrapped

    @classmethod
    def get_credentials(cls):
        if not cls.IS_AUTHENTICATED:
            flgs_args = []
            if GOOGLE_API_AUTH_LOCAL_WEBSERVER:
                flgs_args.append("--noauth_local_webserver")
            flags = argparser.parse_args(flgs_args)
            credential_dir = os.getcwd()
            if not GOOGLE_API_AUTH_LOCAL_CREDENTIAL:
                home_dir = os.path.expanduser("~")
                credential_dir = os.path.join(home_dir, ".credentials")
                if not os.path.exists(credential_dir):
                    os.makedirs(credential_dir)
            credential_path = os.path.join(
                credential_dir, "sheets.googleapis.com-python-mammut.json"
            )
            store = Storage(credential_path)
            credentials = store.get()

            secret_file_name = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), cls.CLIENT_SECRET_FILE
            )
            service_account_path = os.path.join(
                credential_dir, cls.SERVICE_ACCOUNT_SECRET_FILE
            )

            if not credentials or credentials.invalid:
                if os.path.isfile(service_account_path):
                    print("service account exist, google_storage_sheet.py")
                    credentials = service_account.ServiceAccountCredentials.from_json_keyfile_name(
                        service_account_path, scopes=cls.SCOPES
                    )
                    store.locked_put(credentials)
                else:
                    print("service account not found, google_storage_sheet.py")
                    flow = client.flow_from_clientsecrets(
                        secret_file_name,
                        cls.SCOPES,
                        redirect_uri="urn:ietf:wg:oauth:2.0:oob",
                    )
                    flow.user_agent = cls.APPLICATION_NAME
                    credentials = tools.run_flow(flow, store, flags)

            http = credentials.authorize(httplib2.Http())
            discoveryUrl = "https://sheets.googleapis.com/$discovery/rest?" "version=v4"
            cls.SHEETS_SERVICE = discovery.build(
                "sheets", "v4", http=http, discoveryServiceUrl=discoveryUrl
            )
            cls.DRIVE_SERVICE = discovery.build("drive", "v3", http=http)
            cls.SLIDES_SERVICE = discovery.build("slides", "v1", http=http)
            cls.CREDENTIALS = credentials
            cls.IS_AUTHENTICATED = True

        return cls.CREDENTIALS

    @classmethod
    def _get_next_diff(cls, op_type: GoogleAPIOpType) -> int:
        next_in_list = cls.EXECUTIONS_TIME_LIST[op_type][0]
        current_time = mammut.get_timezone_aware_datetime_now()
        diff = current_time - next_in_list
        current_diff_sec = diff.total_seconds()
        return current_diff_sec

    @classmethod
    def _get_next_sleep(cls, op_type: GoogleAPIOpType) -> float:
        next_in_list = cls.EXECUTIONS_TIME_LIST[op_type][0]
        if len(cls.EXECUTIONS_TIME_LIST[op_type]) >= 1:
            next_next_in_list = cls.EXECUTIONS_TIME_LIST[op_type][1]
        else:
            next_next_in_list = mammut.get_timezone_aware_datetime_now()
        diff = next_next_in_list - next_in_list
        next_diff_sec = diff.total_seconds()
        if float(next_diff_sec) >= cls.MIN_API_OP_DELAY:
            return next_diff_sec
        else:
            return cls.MIN_API_OP_DELAY

    @classmethod
    def _check_exec_limit(cls, op_type: GoogleAPIOpType):
        if len(cls.EXECUTIONS_TIME_LIST[op_type]) >= cls.EXECUTION_LIMIT[op_type]:
            while cls._get_next_diff(op_type) <= cls.EXECUTION_TIME_LIMIT[op_type]:
                current_diff_sec = cls._get_next_diff(op_type)
                # If we met the read quota limit, we need to wait at least this
                # difference
                match_quata_limit_sleep = (
                    cls.EXECUTION_TIME_LIMIT[op_type] - current_diff_sec
                )
                total_next_sleep = (
                    cls._get_next_sleep(op_type) + match_quata_limit_sleep
                )
                logger.debug(f"Current next diff {current_diff_sec} - sleeping")
                sleep(total_next_sleep)
                cls.EXECUTIONS_TIME_LIST[op_type] = cls.EXECUTIONS_TIME_LIST[op_type][
                    1:
                ]


class GoogleStorageDrive(GoogleAPI):
    @classmethod
    @GoogleAPI.ensure_credentials_and_read_limits
    def exist_folder(cls, folder_id):
        cls.get_credentials()

        folder_found = True
        try:
            result_images_desc = (
                GoogleStorageDrive.DRIVE_SERVICE.files().get(fileId=folder_id).execute()
            )
            if not result_images_desc.get("id", ""):
                folder_found = False
        except:
            folder_found = False
        return folder_found

    @classmethod
    @GoogleAPI.ensure_credentials_and_read_limits
    def get_folders_in_folder(cls, folder_id):
        cls.get_credentials()
        files_tuples = list()
        response = (
            GoogleStorageDrive.DRIVE_SERVICE.files()
            .list(
                q=f"'{folder_id}' in parents",
                fields="files(id, name, mimeType, modifiedTime)",
            )
            .execute()
        )
        for elem in response.get("files"):
            if elem["mimeType"] == "application/vnd.google-apps.folder":
                files_tuples.append((elem["name"], elem["id"]))
        return files_tuples

    @classmethod
    @GoogleAPI.ensure_credentials_and_read_limits
    def get_all_files_in_folder(cls, folder_id):
        logger.info(
            "Download files from folder: method-get_all_files_in_folder "
            + str(folder_id)
        )
        cls.get_credentials()
        file_tuples = {}
        file_dates = {}

        def add_tuple(item):
            tuple = (item["name"], item["id"])
            file_tuples[item["name"]] = tuple
            file_dates[item["name"]] = get_datetime_from_rfc3339(item["modifiedTime"])

        result_files = (
            GoogleStorageDrive.DRIVE_SERVICE.files()
            .list(
                q=f"'{folder_id}' in parents",
                fields="files(id, name, mimeType, modifiedTime)",
            )
            .execute()
        )
        items = result_files.get("files", [])
        if items:
            for item in items:
                if item["mimeType"] != "application/vnd.google-apps.folder":
                    if item["name"] not in file_dates:
                        add_tuple(item)
                    else:
                        current_time = file_dates[item["name"]]
                        new_time = get_datetime_from_rfc3339(item["modifiedTime"])
                        if new_time > current_time:
                            add_tuple(item)
        return list(file_tuples.values())

    @classmethod
    @GoogleAPI.ensure_credentials_and_read_limits
    def download_files(
        cls,
        source_folder_id: str,
        dest_path: str = os.path.join(os.path.expanduser("~"), "temp"),
        file_names: List[str] = [],
        show_progress=None,
    ):
        logger.info(
            "Download file from folder: method-download_files" + str(source_folder_id)
        )
        cls.get_credentials()
        if not os.path.exists(dest_path):
            os.makedirs(dest_path)
        file_tuples = []
        result_files = (
            GoogleStorageDrive.DRIVE_SERVICE.files()
            .list(q=f"'{source_folder_id}' in parents")
            .execute()
        )
        items = result_files.get("files", [])
        if items:
            for item in items:
                if item["mimeType"] != "application/vnd.google-apps.folder":
                    if len(file_names) == 0 or item["name"] in file_names:
                        result_file = (
                            GoogleStorageDrive.DRIVE_SERVICE.files()
                            .get(
                                fileId=item["id"],
                                fields="mimeType,name,id,webContentLink,webViewLink",
                            )
                            .execute()
                        )
                        request = GoogleStorageDrive.DRIVE_SERVICE.files().get_media(
                            fileId=item["id"]
                        )
                        full_path = os.path.join(dest_path, item["name"])
                        if not os.path.exists(full_path):
                            fh = io.FileIO(full_path, mode="wb")
                            downloader = MediaIoBaseDownload(fh, request)
                            done = False
                            while done is False:
                                status, done = downloader.next_chunk()
                                if show_progress != None:
                                    show_progress(
                                        "Downloading {} - {}".format(
                                            item["name"], int(status.progress() * 100)
                                        )
                                    )
                            fh.close()
                        preview_url = "https://drive.google.com/file/d/{}/preview".format(
                            result_file["id"]
                        )
                        file_tuples.append(
                            (
                                result_file["name"],
                                result_file["id"],
                                result_file["webContentLink"],
                                preview_url,
                                result_file["webViewLink"],
                            )
                        )
                else:
                    cls.download_files(item["id"], dest_path, file_names, show_progress)
        return file_tuples


NGRAM_CORPORA = {
    "english": 15,
    "american english": 17,
    "british english": 18,
    "english fiction": 16,
    "chinese": 23,
    "french": 19,
    "german": 20,
    "hebrew": 24,
    "italian": 22,
    "russian": 25,
    "spanish": 21,
}

NGRAM_CORPORA_BY_REGIONAL_SETTINGS = {"en": "english", "sp": "spanish", "es": "spanish"}


def get_ngram_db_file_names() -> str:
    file_name = resource_filename("mammut.resources.ngram", "google_ngram_cache.db")
    file_name = file_name[:-3]
    return file_name


@GoogleAPI.ensure_credentials_and_ngram_limits
def ngram_frequency(
    token,
    regional_settings,
    smoothing=0,
    year_start=2000,
    year_end=2008,
    case_insensitive=True,
):
    def inner_func(file_db):
        with dbm.open(file_db, "c") as db:
            token_db_key = "{} <---> {}".format(regional_settings, token)
            if token_db_key in db:
                res_from_db = struct.unpack("f", db[token_db_key])[0]
                return res_from_db
            else:
                res = 0.0
                corpus_id = NGRAM_CORPORA[
                    NGRAM_CORPORA_BY_REGIONAL_SETTINGS[regional_settings]
                ]
                token_url = parse.quote(token)
                case_insen = ""
                if case_insensitive:
                    case_insen = "&case_insensitive=on"
                url = (
                    "https://books.google.com/ngrams/graph?content={}{}&year_start={}&year_end={}"
                    "&corpus={}&smoothing={}".format(
                        token_url,
                        case_insen,
                        year_start,
                        year_end,
                        corpus_id,
                        smoothing,
                    )
                )
                page = request.urlopen(url).read()
                start = page.find("data = ".encode("utf-8"))
                end = page.find("];\n".encode("utf-8"), start)
                data = eval(page[start + 8 : end])
                if not isinstance(data, tuple) and not isinstance(data, list):
                    data = [data]
                for d in data:
                    if "type" in d and d["type"] == "CASE_INSENSITIVE":
                        res = float(d["timeseries"][-1])
                        break
                    else:
                        res = res + float(d["timeseries"][-1])
                res_db = struct.pack("f", res)
                db[token_db_key] = res_db
                return res

    file_name = get_ngram_db_file_names()
    return inner_func(file_name)


def del_cache_ngram_frequency(token, regional_settings):
    def inner_del_func(file_db):
        with dbm.open(file_db, "c") as db:
            token_db_key = "{} <---> {}".format(regional_settings, token)
            if token_db_key in db:
                del db[token_db_key]

    file_name = get_ngram_db_file_names()
    inner_del_func(file_name)


def unigram_frecuency(
    token,
    regional_settings,
    smoothing=0,
    year_start=2000,
    year_end=2008,
    case_insensitive=True,
):
    return ngram_frequency(
        token, regional_settings, smoothing, year_start, year_end, case_insensitive
    )


def unordered_bigram_frecuency(
    tokens_bituple,
    regional_settings,
    smoothing=0,
    year_start=2000,
    year_end=2008,
    case_insensitive=True,
):
    bigram_1 = tokens_bituple[0] + " " + tokens_bituple[1]
    res = ngram_frequency(
        bigram_1, regional_settings, smoothing, year_start, year_end, case_insensitive
    )
    bigram_2 = tokens_bituple[1] + " " + tokens_bituple[0]
    res = res + ngram_frequency(
        bigram_2, regional_settings, smoothing, year_start, year_end, case_insensitive
    )
    return res


def unordered_trigram_frecuency(
    tokens_trituple,
    regional_settings,
    smoothing=0,
    year_start=2000,
    year_end=2008,
    case_insensitive=True,
):
    trigram_123 = (
        tokens_trituple[0] + " " + tokens_trituple[1] + " " + tokens_trituple[2]
    )
    res = ngram_frequency(
        trigram_123,
        regional_settings,
        smoothing,
        year_start,
        year_end,
        case_insensitive,
    )
    trigram_132 = (
        tokens_trituple[0] + " " + tokens_trituple[2] + " " + tokens_trituple[1]
    )
    res = res + ngram_frequency(
        trigram_132,
        regional_settings,
        smoothing,
        year_start,
        year_end,
        case_insensitive,
    )
    trigram_213 = (
        tokens_trituple[1] + " " + tokens_trituple[0] + " " + tokens_trituple[2]
    )
    res = res + ngram_frequency(
        trigram_213,
        regional_settings,
        smoothing,
        year_start,
        year_end,
        case_insensitive,
    )
    trigram_231 = (
        tokens_trituple[1] + " " + tokens_trituple[2] + " " + tokens_trituple[0]
    )
    res = res + ngram_frequency(
        trigram_231,
        regional_settings,
        smoothing,
        year_start,
        year_end,
        case_insensitive,
    )
    trigram_312 = (
        tokens_trituple[2] + " " + tokens_trituple[0] + " " + tokens_trituple[1]
    )
    res = res + ngram_frequency(
        trigram_312,
        regional_settings,
        smoothing,
        year_start,
        year_end,
        case_insensitive,
    )
    trigram_321 = (
        tokens_trituple[2] + " " + tokens_trituple[1] + " " + tokens_trituple[0]
    )
    res = res + ngram_frequency(
        trigram_321,
        regional_settings,
        smoothing,
        year_start,
        year_end,
        case_insensitive,
    )
    return res



from typing import List, Dict
from mammut.curriculum.course.course import (
    Course,
    KttPackageInterpretationCourse,
)
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
import mammut.curriculum.core.utils as curriculum_utils
import mammut.curriculum.report as report
import networkx as nx


class Flowchart:
    """Encapsulates graph structure for the curriculum shape.

        Class invariants:
            - The curriculum flowchart is a directed acyclic graph.
            - Isolated courses are not allowed.
            - Courses ID's are only allowed to be integers.
    """

    def __init__(self, curriculum: "CurriculumBase"):
        """
        Args:
            curriculum: Curriculum object to get references from.
        """
        self.graph = nx.DiGraph()
        self.graph.add_nodes_from(curriculum.courses)
        self.valid = True
        self.errors: List[str] = []
        """
            Creates a local dictionary for id->node(course object).
            Used in second iteration to add the directed edge from a prerequisite 
            node to the prelate node. 
        """
        self.graph_dict: Dict = {}
        for course in curriculum.courses:
            self.graph_dict[course.course_id] = course
        for course in curriculum.courses:
            for prerequisite in course.prerequisites:
                self.graph.add_edge(self.graph_dict[prerequisite], course)

        """
            Perform validations to the flowchart shape. 
            If an error is found, flowchart is marked invalid, 
            and errors can be requested in curriculum client class.
        """
        if not nx.algorithms.isolate.number_of_isolates(self.graph) == 0:
            self.valid = False
            error_message_str = f"""Curriculum contains isolated courses {
                [course.course_id for course in nx.algorithms.isolate.isolates(self.graph)]
                }"""
            report.send_message(
                report.CurriculumGeneralErrorMessage(error_message_str)
            )
            self.errors.append(error_message_str)

        if not nx.algorithms.dag.is_directed_acyclic_graph(self.graph):
            self.valid = False
            error_message_str = "Curriculum flowchart contains cycles."
            report.send_message(
                report.CurriculumGeneralErrorMessage(error_message_str)
            )
            self.errors.append(error_message_str)

    def add_ktt_interpretation_course(self, curriculum: "CurriculumBase"):
        """Adds the KtT compilation interpretation course as the first course
        of the curriculum, in the academic period 0.

        It also adds the course as prerequisite to all the courses from the academic
        period 1.
        """
        ktt_package_course = KttPackageInterpretationCourse(
            curriculum._mammut_session_context, curriculum.assessor
        )
        curriculum.courses.insert(0, ktt_package_course)
        first_academic_period_courses = [
            course
            for course in curriculum.courses
            if course.academic_period == 1
        ]
        for course in first_academic_period_courses:
            course.prerequisites = [0]

        self.graph_dict[ktt_package_course.course_id] = ktt_package_course
        for course in first_academic_period_courses:
            self.graph.add_edge(
                self.graph_dict[ktt_package_course.course_id], course
            )

    def get_max_parallel_courses(self) -> int:
        """Returns the maximum number of parallel courses in the academic
        periods.
        """
        academic_periods_table: Dict = {}
        for c in self.graph.nodes:
            if c.academic_period not in academic_periods_table:
                academic_periods_table[c.academic_period] = 1
            else:
                academic_periods_table[c.academic_period] += 1

        return max(academic_periods_table.values())





class FeatureDescriptor:
    def __init__(
        self,
        feature_class_group: str,
        feature_class: str,
        feature_name: str,
        feature_id: str,
        pos_tags: str,
        documentation_link: str,
        statistics_link: str,
        standard,
    ):
        self.feature_class_group = feature_class_group
        self.feature_class = feature_class
        self.feature_name = feature_name
        self.feature_id = feature_id
        self.pos_tags = [pt.strip() for pt in pos_tags.split(",")]
        self.documentation_link = documentation_link
        self.statistics_link = statistics_link
        self.standard = standard

    def __eq__(self, other):
        if isinstance(other, FeatureDescriptor):
            return (
                self.feature_class_group == other.feature_class_group
                and self.feature_class == other.feature_class
                and self.feature_name == other.feature_name
                and self.feature_id == other.feature_id
                and self.documentation_link == other.documentation_link
                and self.statistics_link == other.statistics_link
            )
        return False

    def __str__(self):
        return f"{self.feature_class}:{self.feature_name} ({self.feature_id})"



from mammut.assessment.problem import Problem
from mammut.curriculum.models.sentencepiece import SentencepieceBaseModel
from typing import List, Tuple, Dict


class SpecialTokenProblemMFT(Problem):
    """This class evaluates the encoding produced by a trained
    tokenizer.

    Test the minimal functionality: expected token(s) is(are) found in the
    encoded output.
    """

    def __init__(
        self,
        course_id: str,
        assessment_id: str,
        schedule: str,
        problem_id: str,
        problem_type: str,
        parameters: str,
        language_competence: str,
        weight: str,
        retry: str,
        observations: str,
    ):
        """
        Args:
            course_id(str): id of the course to which this problem belongs.
            assessment_id(str): id of the assessment that bundles this problem.
            schedule(str): lesson schedule to execute this problem.
            problem_id(str): id for this problem.
            problem_type(str): problem type enumeration string.
            parameters (str): JSON string parameters for this problem.
            language_competence(str): language competence enumeration value for this problem.
            weight(str): float string value for the weight of this particular problem
                in the overall course score.
            observations(str): observations string. Only descriptive.
        """
        Problem.__init__(
            self,
            course_id,
            assessment_id,
            schedule,
            problem_id,
            problem_type,
            parameters,
            language_competence,
            weight,
            retry,
            observations,
        )
        self._input: str = self.parameters["input"]
        self._expected_tokens_dict = self.parameters["expected_tokens"]
        self._tokenized_output: str = ""

    def solve(self, model: SentencepieceBaseModel, **kwargs) -> bool:
        self._retries_count += 1
        subword_tokenizer = model.sp_preprocessor
        self._tokenized_output = subword_tokenizer.encode(
            self._input, out_type=str
        )
        for (
            expected_token,
            expected_occurences,
        ) in self._expected_tokens_dict.items():
            if (
                not self._tokenized_output.count(expected_token)
                == expected_occurences
            ):
                """
                    If count does not match, short circuit and return false.
                """
                return False

        return True

    def get_result_info(self) -> Dict:
        info = {
            "Tokenization Result": ", ".join(self._tokenized_output),
            "Expected tokens": self._expected_tokens_dict,
        }
        return info



# coding=utf-8
from datetime import datetime
import logging

import pandas as pd
from googleapiclient.errors import HttpError
import mammut
from mammut.common.google_api import GoogleAPI
from mammut.common.storage.storage_base import SheetsStorageBase

logger = logging.getLogger()


class GoogleStorageSheet(SheetsStorageBase, GoogleAPI):
    DEFAULT_RANGE_SEPARATOR = "!"

    def __init__(self):
        SheetsStorageBase.__init__(self)

    @GoogleAPI.ensure_credentials_and_read_limits
    def get_spreadsheet_as_dataframe(
        self,
        spreadsheet_id: str,
        sheet_name,
        start_column: str,
        start_row: str,
        end_range: str,
        transpose: bool = False,
    ):
        range_name_start = (
            sheet_name
            + GoogleStorageSheet.DEFAULT_RANGE_SEPARATOR
            + start_column
            + start_row
        )
        range_data = range_name_start + ((":" + end_range) if end_range else "")
        majorDimension = "ROWS" if not transpose else "COLUMNS"

        def check_row_size(row, size):
            res = row
            if len(row) < size:
                res = row + ["" for i in range(0, size - len(row))]
            return res

        result = (
            GoogleStorageSheet.SHEETS_SERVICE.spreadsheets()
            .values()
            .get(
                spreadsheetId=spreadsheet_id,
                range=range_data,
                majorDimension=majorDimension,
            )
            .execute()
        )
        result_values = result.get("values", [])
        col_names = [h.strip() for h in result_values[0]]
        data = [check_row_size(r, len(col_names)) for r in result_values[1:]]
        data_frame = pd.DataFrame(data, columns=col_names)
        return data_frame

    @GoogleAPI.ensure_credentials_and_write_limits
    def add_rows_to_spreadsheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        rows,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        end_range=SheetsStorageBase.DEFAULT_END_RANGE,
    ):
        range_name_start = (
            sheet_name
            + GoogleStorageSheet.DEFAULT_RANGE_SEPARATOR
            + start_col
            + start_row
        )
        body = {"values": rows}
        result = (
            GoogleStorageSheet.SHEETS_SERVICE.spreadsheets()
            .values()
            .append(
                spreadsheetId=spreadsheet_id,
                range=range_name_start,
                valueInputOption="RAW",
                body=body,
            )
            .execute()
        )
        return result.get("tableRange")

    @GoogleAPI.ensure_credentials_and_write_limits
    def append_row_to_sheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        row,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        end_range=SheetsStorageBase.DEFAULT_END_RANGE,
    ):
        value_input_option = "RAW"
        value_range_body = {"values": [row]}

        request = (
            GoogleStorageSheet.SHEETS_SERVICE.spreadsheets()
            .values()
            .append(
                spreadsheetId=spreadsheet_id,
                range=sheet_name,
                valueInputOption=value_input_option,
                body=value_range_body,
            )
        )
        response = request.execute()
        return response

    @GoogleAPI.ensure_credentials_and_write_limits
    def create_new_sheet(self, spreadsheet_id: str, sheet_name: str):
        requests = []
        requests.append({"addSheet": {"properties": {"title": sheet_name}}})
        body = {"requests": requests}
        return (
            GoogleStorageSheet.SHEETS_SERVICE.spreadsheets()
            .batchUpdate(spreadsheetId=spreadsheet_id, body=body)
            .execute()
        )

    @GoogleAPI.ensure_credentials_and_write_limits
    def create_temporal_new_sheet(self, spreadsheet_id: str, base_sheet_title: str):
        sheet_title = (
            base_sheet_title + "-" + mammut.get_str_format_timezone_aware_datetime_now()
        )
        sheet_title = sheet_title.replace(
            mammut.TIME_SEPARATOR, mammut.TIME_ALTERNATIVE_SEPARATOR
        )
        self.create_new_sheet(spreadsheet_id, sheet_title)
        return sheet_title

    @GoogleAPI.ensure_credentials_and_read_limits
    def get_temporal_sheets_from_new_to_old(self, spreadsheet_id, base_sheet_title):
        request = GoogleStorageSheet.SHEETS_SERVICE.spreadsheets().get(
            spreadsheetId=spreadsheet_id
        )
        response = request.execute()
        result = []
        for s in response["sheets"]:
            title = s["properties"]["title"]
            if base_sheet_title in title and len(base_sheet_title) != len(title):
                time_stamp_str = title[len(base_sheet_title) + 1 :]
                time_stamp_str = time_stamp_str.replace(
                    mammut.TIME_ALTERNATIVE_SEPARATOR, mammut.TIME_SEPARATOR
                )
                time_stamp = mammut.str_parse_datetime(time_stamp_str)
                result.append((s["properties"]["sheetId"], title, time_stamp))
        result = sorted(result, key=lambda i: i[2], reverse=True)
        return result

    @GoogleAPI.ensure_credentials_and_read_limits
    def get_sheet_id(self, spreadsheet_id, identifier):
        request = GoogleStorageSheet.SHEETS_SERVICE.spreadsheets().get(
            spreadsheetId=spreadsheet_id, ranges=identifier
        )
        response = request.execute()
        result = response["sheets"][0]["properties"]["sheetId"]
        return result

    @GoogleAPI.ensure_credentials_and_write_limits
    def update_row_from_spreadsheet(
        self, spreadsheet_id, sheet_name, row, start_row, start_col
    ):
        start_index = (
            sheet_name
            + GoogleStorageSheet.DEFAULT_RANGE_SEPARATOR
            + start_col
            + start_row
        )
        value_input_option = "RAW"
        value_range_body = {"values": [row]}
        request = (
            GoogleStorageSheet.SHEETS_SERVICE.spreadsheets()
            .values()
            .update(
                spreadsheetId=spreadsheet_id,
                range=start_index,
                valueInputOption=value_input_option,
                body=value_range_body,
            )
        )
        response = request.execute()
        return response

    @GoogleAPI.ensure_credentials_and_write_limits
    def update_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, rows, start_row, start_col
    ):
        start_index = (
            sheet_name
            + GoogleStorageSheet.DEFAULT_RANGE_SEPARATOR
            + start_col
            + start_row
        )
        value_input_option = "RAW"
        value_range_body = {"values": rows}
        request = (
            GoogleStorageSheet.SHEETS_SERVICE.spreadsheets()
            .values()
            .update(
                spreadsheetId=spreadsheet_id,
                range=start_index,
                valueInputOption=value_input_option,
                body=value_range_body,
            )
        )
        response = request.execute()
        return response

    @GoogleAPI.ensure_credentials_and_write_limits
    def delete_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, start_index, end_index
    ):
        sheet_id = self.get_sheet_id(spreadsheet_id, sheet_name)
        batch_update_spreadsheet_request_body = {
            "requests": [
                {
                    "deleteDimension": {
                        "range": {
                            "sheetId": sheet_id,
                            "dimension": "ROWS",
                            "startIndex": start_index,  # desde sin incluir
                            "endIndex": end_index,  # hasta, incluyendolo
                        }
                    }
                }
            ]
        }
        request = GoogleStorageSheet.SHEETS_SERVICE.spreadsheets().batchUpdate(
            spreadsheetId=spreadsheet_id, body=batch_update_spreadsheet_request_body
        )
        response = request.execute()
        return response

    @GoogleAPI.ensure_credentials_and_write_limits
    def delete_sheet(self, spreadsheet_id, sheet_name):
        sheet_id = self.get_sheet_id(spreadsheet_id, sheet_name)
        batch_update_spreadsheet_request_body = {
            "requests": [{"deleteSheet": {"sheetId": sheet_id,}}]
        }
        request = GoogleStorageSheet.SHEETS_SERVICE.spreadsheets().batchUpdate(
            spreadsheetId=spreadsheet_id, body=batch_update_spreadsheet_request_body
        )
        response = request.execute()
        return response



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mammut',
 'mammut.apps',
 'mammut.assessment',
 'mammut.common',
 'mammut.common.catalog',
 'mammut.common.corpus',
 'mammut.common.lexicon',
 'mammut.common.storage',
 'mammut.common.synthetic',
 'mammut.common.synthetic.synthetic_corpus_generator',
 'mammut.curriculum',
 'mammut.curriculum.classroom',
 'mammut.curriculum.core',
 'mammut.curriculum.course',
 'mammut.curriculum.lesson',
 'mammut.curriculum.models',
 'mammut.curriculum.models.hugging_face',
 'mammut.curriculum.models.online_models',
 'mammut.curriculum.report',
 'mammut.curriculum.report.errors',
 'mammut.curriculum.report.messages',
 'mammut.curriculum.report.reporters',
 'mammut.data',
 'mammut.linguistics',
 'mammut.mammutctl',
 'mammut.models',
 'mammut.models.imaging',
 'mammut.models.reparameterization',
 'mammut.resources',
 'mammut.resources.ngram',
 'mammut.visualization',
 'mammut.visualization.templates']

package_data = \
{'': ['*']}

install_requires = \
['Augmentor>=0.2.8,<0.3.0',
 'brat-widget>=0.2.5,<0.3.0',
 'chart-studio>=1.1.0,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'datasets>=1.2.1,<2.0.0',
 'django>=3.0.7,<4.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'elasticsearch>=7.7.1,<8.0.0',
 'google-api-python-client>=1.9.3,<2.0.0',
 'graphviz>=0.14,<0.15',
 'jupyter>=1.0.0,<2.0.0',
 'kafka-python>=2.0.1,<3.0.0',
 'networkx>=2.5,<3.0',
 'nltk>=3.5,<4.0',
 'oauth2client>=4.1.3,<5.0.0',
 'openpyxl>=3.0.3,<4.0.0',
 'packaging>=20.4,<21.0',
 'pandas>=1.0.4,<2.0.0',
 'parsec>=3.5,<4.0',
 'pip>=20.0.0,<21.0.0',
 'plotly>=4.8.1,<5.0.0',
 'pytest-order>=0.10.0,<0.11.0',
 'python-pptx>=0.6.18,<0.7.0',
 'ray>=1.2.0,<2.0.0',
 'redis>=3.5.3,<4.0.0',
 'scikit-learn>=0.23.1,<0.24.0',
 'sentencepiece>=0.1.92,<0.2.0',
 'simpleneighbors>=0.1.0,<0.2.0',
 'tensorflow-probability==0.10.0',
 'tensorflow==2.2.0',
 'tensorflow_hub[make_image_classifier]==0.8.0',
 'tf-sentencepiece>=0.1.90,<0.2.0',
 'torch>=1.8.1,<2.0.0',
 'transformers>=4.5.1,<5.0.0',
 'transitions>=0.8.5,<0.9.0',
 'xlrd>=1.2.0,<2.0.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses==0.6']}

entry_points = \
{'console_scripts': ['mammutctl = src.mammut.mammutctl.mammutctl:main']}

setup_kwargs = {
    'name': 'mammut-py',
    'version': '0.1.0.dev202110120327',
    'description': 'Mammut framework is an open library for computational linguistics.',
    'long_description': '# Mammut-Py\n\n## Installation\n\nUse pip to install the package:\n```\npip install mammut-py\n```\n\n## Development\n\nThe following tools are used in this project:\n- [Poetry](https://python-poetry.org/) is used as package manager.\n- [Nox](https://nox.thea.codes/) is used as automation tool, mainly for testing.\n- [Black](https://black.readthedocs.io/) is the mandatory formatter tool.\n- [PyEnv](https://github.com/pyenv/pyenv/wiki) is recommended as a tool to handle multiple python versions in your machine.\n\nThe library is intended to be compatible with python ~3.6.9, ~3.7.4 and ~3.8.2. But the primary version to support is ~3.8.2.\n\nThe general structure of the project is trying to follow the recommendations\nin [Cookiecutter Data Science](https://drivendata.github.io/cookiecutter-data-science/).\nThe main difference lies in the source code itself which is not constraint to data science code.\n',
    'author': 'Mammut.io',
    'author_email': 'support@mammut.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://doc.mammut.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import dataclasses
import ray
import datetime
from typing import Optional
from mammut import config


@dataclasses.dataclass
class MammutSessionContext:
    """Data class to encapsulate the mammut curriculum context
    data across different processes.

    Currently, this class should encapsulate values that can be serialized
    out-of-the-box by Ray.


    TODO: there are details to work on. Right now is being used
        for main report registry actor handle propagation. But
        a good conceptual design is required.
    """

    package_id: str
    main_report_actor_handle: Optional[ray.actor.ActorHandle] = None
    session_id: str = str(datetime.datetime.now())
    mammut_id: int = 1
    ray_classroom_id: int = 0
    curriculum_save_folder: str = config.save_folder
    online_models_registry: Optional[ray.actor.ActorHandle] = None



# coding=utf-8
from mammut.common.simulator_utilities import *
from ipywidgets import *

### Room Layout ###
def get_room_form():
    form_values = dict()
    roomDropdown = widgets.Dropdown(
        description="Room:", disabled=False, button_style=""
    )
    form_values["roomDropdown"] = roomDropdown

    newRoomLayout = Layout(width="100%", height="50px")
    nameTextBox = widgets.Text(
        value="", placeholder="test room", description="Nombre:", layout=newRoomLayout
    )
    form_values["nameTextBox"] = nameTextBox
    roomTypeDropdown = widgets.Dropdown(
        description="Room Type:", disabled=False, button_style=""
    )
    room_type_values_dict = {room_type.value: room_type.value for room_type in RoomType}

    roomTypeDropdown.options = room_type_values_dict
    form_values["roomTypeDropdown"] = roomTypeDropdown

    newRoomButton = widgets.Button(
        description="Crear Nuevo ...", button_style="primary"
    )

    roomBox = widgets.VBox(
        [
            widgets.HBox(
                [
                    widgets.VBox(
                        [nameTextBox, roomTypeDropdown], layout=Layout(width="100%")
                    )
                ],
                layout=Layout(width="100%"),
            ),
            newRoomButton,
        ]
    )

    selectRoomButton = widgets.Button(description="Seleccionar", button_style="primary")

    selectRoomBox = widgets.HBox([roomDropdown, selectRoomButton])

    room_tab = widgets.Tab(children=[selectRoomBox, roomBox])
    room_tab.set_title(0, "Seleccionar Room")
    room_tab.set_title(1, "Crear Nuevo Room...")

    return selectRoomButton, newRoomButton, room_tab, form_values



# coding=utf-8

from collections import *
import itertools

spanish_venezuela_code = 'sp-ve'
base_free_morpheme_type = 'Base-Free'

lexical_def_type = 'Lexical'
limited_def_type = 'Limited'
definition_types = [lexical_def_type, limited_def_type]

verb_pos_tag = 'VERB'
noun_pos_tag = 'NOUN'
propn_pos_tag = 'PROPN'
adj_pos_tag = 'ADJ'
adv_pos_tag = 'ADV'
intj_pos_tag = 'INTJ'
det_pos_tag = 'DET'
adp_pos_tag = 'ADP'
cconj_pos_tag = 'CCONJ'
sconj_pos_tag = 'SCONJ'
part_pos_tag = 'PART'
pron_pos_tag = 'PRON'
aux_pos_tag = 'AUX'
num_pos_tag = 'NUM'
punct_pos_tag = 'PUNCT'
sym_pos_tag = 'SYM'
x_pos_tag = 'X'
pos_tags = [verb_pos_tag,noun_pos_tag,propn_pos_tag,adj_pos_tag,adv_pos_tag,intj_pos_tag,det_pos_tag,adp_pos_tag,cconj_pos_tag,
            sconj_pos_tag,part_pos_tag,pron_pos_tag,aux_pos_tag,num_pos_tag,punct_pos_tag,sym_pos_tag,x_pos_tag]
pos_tags_description_sites = {}
pos_tags_description_sites[verb_pos_tag] = 'http://universaldependencies.org/u/pos/VERB.html'
pos_tags_description_sites[noun_pos_tag] = 'http://universaldependencies.org/u/pos/NOUN.html'
pos_tags_description_sites[propn_pos_tag] = 'http://universaldependencies.org/u/pos/PROPN.html'
pos_tags_description_sites[adj_pos_tag] = 'http://universaldependencies.org/u/pos/ADJ.html'
pos_tags_description_sites[adv_pos_tag] = 'http://universaldependencies.org/u/pos/ADV.html'
pos_tags_description_sites[intj_pos_tag] = 'http://universaldependencies.org/u/pos/INTJ.html'
pos_tags_description_sites[det_pos_tag] = 'http://universaldependencies.org/u/pos/DET.html'
pos_tags_description_sites[adp_pos_tag] = 'http://universaldependencies.org/u/pos/ADP.html'
pos_tags_description_sites[cconj_pos_tag] = 'http://universaldependencies.org/u/pos/CCONJ.html'
pos_tags_description_sites[sconj_pos_tag] = 'http://universaldependencies.org/u/pos/SCONJ.html'
pos_tags_description_sites[part_pos_tag] = 'http://universaldependencies.org/u/pos/PART.html'
pos_tags_description_sites[pron_pos_tag] = 'http://universaldependencies.org/u/pos/PRON.html'
pos_tags_description_sites[aux_pos_tag] = 'http://universaldependencies.org/u/pos/AUX_.html'
pos_tags_description_sites[num_pos_tag] = 'http://universaldependencies.org/u/pos/NUM.html'
pos_tags_description_sites[punct_pos_tag] = 'http://universaldependencies.org/u/pos/PUNCT.html'
pos_tags_description_sites[sym_pos_tag] = 'http://universaldependencies.org/u/pos/SYM.html'
pos_tags_description_sites[x_pos_tag] = 'http://universaldependencies.org/u/pos/X.html'
pos_tags_stats_sites_es = {}
pos_tags_stats_sites_es[verb_pos_tag] = 'http://universaldependencies.org/es/pos/VERB.html'
pos_tags_stats_sites_es[noun_pos_tag] = 'http://universaldependencies.org/es/pos/NOUN.html'
pos_tags_stats_sites_es[propn_pos_tag] = 'http://universaldependencies.org/es/pos/PROPN.html'
pos_tags_stats_sites_es[adj_pos_tag] = 'http://universaldependencies.org/es/pos/ADJ.html'
pos_tags_stats_sites_es[adv_pos_tag] = 'http://universaldependencies.org/es/pos/ADV.html'
pos_tags_stats_sites_es[intj_pos_tag] = 'http://universaldependencies.org/es/pos/INTJ.html'
pos_tags_stats_sites_es[det_pos_tag] = 'http://universaldependencies.org/es/pos/DET.html'
pos_tags_stats_sites_es[adp_pos_tag] = 'http://universaldependencies.org/es/pos/ADP.html'
pos_tags_stats_sites_es[cconj_pos_tag] = 'http://universaldependencies.org/es/pos/CONJ.html'
pos_tags_stats_sites_es[sconj_pos_tag] = 'http://universaldependencies.org/es/pos/SCONJ.html'
pos_tags_stats_sites_es[part_pos_tag] = 'http://universaldependencies.org/es/pos/PART.html'
pos_tags_stats_sites_es[pron_pos_tag] = 'http://universaldependencies.org/es/pos/PRON.html'
pos_tags_stats_sites_es[aux_pos_tag] = 'http://universaldependencies.org/es/pos/AUX_.html'
pos_tags_stats_sites_es[num_pos_tag] = 'http://universaldependencies.org/es/pos/NUM.html'
pos_tags_stats_sites_es[punct_pos_tag] = 'http://universaldependencies.org/es/pos/PUNCT.html'
pos_tags_stats_sites_es[sym_pos_tag] = 'http://universaldependencies.org/es/pos/SYM.html'
pos_tags_stats_sites_es[x_pos_tag] = 'http://universaldependencies.org/es/pos/X.html'

FeatureDescriptor = namedtuple('FeatureDescriptor', ['id', 'name', 'description', 'url_doc', 'url_stats_es', 'values'])
feature_descriptor_per_feature = {}
word_features_per_pos_tag = {}
for tag in pos_tags:
    word_features_per_pos_tag[tag] = []

feature_types_descriptor_per_id = {}

pronominal_types_descriptor = FeatureDescriptor('PronType', 'pronominal type',
                                                'applies to pronouns, pronominal adjectives (determiners), pronominal numerals (quantifiers) and pronominal adverbs.',
                                                'http://universaldependencies.org/u/feat/PronType.html', 'http://universaldependencies.org/es/feat/PronType.html',
                                                ['PronType_Prs', 'PronType_Rcp', 'PronType_Art', 'PronType_Int', 'PronType_Rel', 'PronType_Exc',
                                                 'PronType_Dem', 'PronType_Emp', 'PronType_Tot', 'PronType_Neg', 'PronType_Ind'])
for f in pronominal_types_descriptor.values:
    feature_descriptor_per_feature[f] = pronominal_types_descriptor
word_features_per_pos_tag[propn_pos_tag].append(pronominal_types_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(pronominal_types_descriptor.values)
word_features_per_pos_tag[num_pos_tag].append(pronominal_types_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(pronominal_types_descriptor.values)
word_features_per_pos_tag[pron_pos_tag].append(pronominal_types_descriptor.values)
feature_types_descriptor_per_id[pronominal_types_descriptor.id] = pronominal_types_descriptor


numeral_types_descriptor = FeatureDescriptor('NumType', 'numeral type',
                                             'applies to [NUM: cardinal numerals], [DET: quantifiers], [ADJ: definite adjectival, e.g. ordinal numerals], [ADV: adverbial (e.g. ordinal and multiplicative) numerals, both definite and pronominal]',
                                             'http://universaldependencies.org/u/feat/NumType.html', 'http://universaldependencies.org/es/feat/NumType.html',
                                             ['NumType_Card','NumType_Dist','NumType_Frac','NumType_Mult','NumType_Ord','NumType_Range',
                                              'NumType_Sets'])
for f in numeral_types_descriptor.values:
    feature_descriptor_per_feature[f] = numeral_types_descriptor
word_features_per_pos_tag[num_pos_tag].append(numeral_types_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(numeral_types_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(numeral_types_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(numeral_types_descriptor.values)
word_features_per_pos_tag[pron_pos_tag].append(numeral_types_descriptor.values)
feature_types_descriptor_per_id[numeral_types_descriptor.id] = numeral_types_descriptor


possessives_descriptor = FeatureDescriptor('Poss', 'possessive',
                                           'Boolean feature of pronouns, determiners or adjectives. It tells whether the word is possessive.',
                                           'http://universaldependencies.org/u/feat/Poss.html', 'http://universaldependencies.org/es/feat/Poss.html',
                                           ['Poss_Yes'])
for f in possessives_descriptor.values:
    feature_descriptor_per_feature[f] = possessives_descriptor
word_features_per_pos_tag[pron_pos_tag].append(possessives_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(possessives_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(possessives_descriptor.values)
feature_types_descriptor_per_id[possessives_descriptor.id] = possessives_descriptor


reflexives_descriptor = FeatureDescriptor('Reflex', 'reflexive',
                                          'Boolean feature, typically of pronouns or determiners. It tells whether the word is reflexive, i.e. refers to the subject of its clause.',
                                          'http://universaldependencies.org/u/feat/Reflex.html', 'http://universaldependencies.org/es/feat/Reflex.html',
                                          ['Reflex_Yes'])
for f in reflexives_descriptor.values:
    feature_descriptor_per_feature[f] = reflexives_descriptor
word_features_per_pos_tag[pron_pos_tag].append(reflexives_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(reflexives_descriptor.values)
feature_types_descriptor_per_id[reflexives_descriptor.id] = reflexives_descriptor


foreigns_descriptor = FeatureDescriptor('Foreign', 'foreign word',
                                 'Boolean feature. Is this a foreign word? Not a loan word and not a foreign name but a genuinely foreign word appearing inside native text, e.g. inside direct speech, titles of books etc. This feature would apply either to the X part of speech (unanalyzable token), or to other parts of speech',
                                 'http://universaldependencies.org/u/feat/Foreign.html', 'http://universaldependencies.org/es/feat/Foreign.html',
                                 ['Foreign_Yes'])
for f in foreigns_descriptor.values:
    feature_descriptor_per_feature[f] = foreigns_descriptor
word_features_per_pos_tag[x_pos_tag].append(foreigns_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(foreigns_descriptor.values)
word_features_per_pos_tag[verb_pos_tag].append(foreigns_descriptor.values)
word_features_per_pos_tag[propn_pos_tag].append(foreigns_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(foreigns_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(foreigns_descriptor.values)
feature_types_descriptor_per_id[foreigns_descriptor.id] = foreigns_descriptor


abbreviations_descriptor = FeatureDescriptor('Abbr', 'abbreviation word',
                                 'Boolean feature. Is this an abbreviation? Note that the abbreviated word(s) typically belongs to a part of speech other than X.',
                                 'http://universaldependencies.org/u/feat/Abbr.html', 'http://universaldependencies.org/u/feat/Abbr.html',
                                 ['Abbr_Yes'])
for f in abbreviations_descriptor.values:
    feature_descriptor_per_feature[f] = abbreviations_descriptor
word_features_per_pos_tag[x_pos_tag].append(abbreviations_descriptor.values)
feature_types_descriptor_per_id[abbreviations_descriptor.id] = abbreviations_descriptor


genders_descriptor = FeatureDescriptor('Gender', 'gender',
                                 'Gender is usually a lexical feature of nouns and inflectional feature of other parts of speech (pronouns, adjectives, determiners, numerals, verbs) that mark agreement with nouns. ',
                                 'http://universaldependencies.org/u/feat/Gender.html', 'http://universaldependencies.org/es/feat/Gender.html',
                                 ['Gender_Com','Gender_Fem','Gender_Masc','Gender_Neut'])
for f in genders_descriptor.values:
    feature_descriptor_per_feature[f] = genders_descriptor
word_features_per_pos_tag[noun_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[propn_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[num_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[verb_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[pron_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[aux_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[x_pos_tag].append(genders_descriptor.values)
word_features_per_pos_tag[sym_pos_tag].append(genders_descriptor.values)
feature_types_descriptor_per_id[genders_descriptor.id] = genders_descriptor


animacys_descriptor = FeatureDescriptor('Animacy', 'animacy',
                                        'NA Spanish. Similarly to Gender (and to the African noun classes), animacy is usually a lexical feature of nouns and inflectional feature of other parts of speech (pronouns, adjectives, determiners, numerals, verbs) that mark agreement with nouns.',
                                        'http://universaldependencies.org/u/feat/Animacy.html', 'http://universaldependencies.org/es/feat/Animacy.html',
                                        ['Animacy_Anim','Animacy_Hum','Animacy_Inan','Animacy_Nhum'])
for f in animacys_descriptor.values:
    feature_descriptor_per_feature[f] = animacys_descriptor
#word_features_per_pos_tag[noun_pos_tag].append(animacys_descriptor.values)
#word_features_per_pos_tag[propn_pos_tag].append(animacys_descriptor.values)
#word_features_per_pos_tag[adj_pos_tag].append(animacys_descriptor.values)
#word_features_per_pos_tag[det_pos_tag].append(animacys_descriptor.values)
#word_features_per_pos_tag[num_pos_tag].append(animacys_descriptor.values)
#word_features_per_pos_tag[verb_pos_tag].append(animacys_descriptor.values)
feature_types_descriptor_per_id[animacys_descriptor.id] = animacys_descriptor


numbers_descriptor = FeatureDescriptor('Number', ' type',
                                       'Number is usually an inflectional feature of nouns and, depending on language, other parts of speech (pronouns, adjectives, determiners, numerals, verbs) that mark agreement with nouns.',
                                       'http://universaldependencies.org/u/feat/Number.html', 'http://universaldependencies.org/es/feat/Number.html',
                                       ['Number_Coll','Number_Count','Number_Dual','Number_Grpa','Number_Grpl','Number_Inv','Number_Pauc',
                                        'Number_Plur','Number_Ptan','Number_Sing','Number_Tri'])
for f in numbers_descriptor.values:
    feature_descriptor_per_feature[f] = numbers_descriptor
word_features_per_pos_tag[noun_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[propn_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[num_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[verb_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[pron_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[aux_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[x_pos_tag].append(numbers_descriptor.values)
word_features_per_pos_tag[sym_pos_tag].append(numbers_descriptor.values)
feature_types_descriptor_per_id[numbers_descriptor.id] = numbers_descriptor


cases_descriptor = FeatureDescriptor('Case', 'case',
                                     'Case is usually an inflectional feature of nouns and, depending on language, other parts of speech (pronouns, adjectives, determiners, numerals, verbs) that mark agreement with nouns. In some tagsets it is also valency feature of adpositions.',
                                     'http://universaldependencies.org/u/feat/Case.html', 'http://universaldependencies.org/es/feat/Case.html',
                                     ['Case_Abs','Case_Acc','Case_Erg','Case_Nom','Case_Abe','Case_Ben','Case_Cau','Case_Cmp','Case_Com','Case_Dat',
                                      'Case_Dis','Case_Equ','Case_Gen','Case_Ins','Case_Par','Case_Tem','Case_Tra','Case_VocAbl','Case_Add',
                                      'Case_Ade','Case_All','Case_Del','Case_Ela','Case_Ess','Case_Ill','Case_Ine','Case_Lat','Case_Loc','Case_Sub',
                                      'Case_Sup','Case_Ter'])
for f in cases_descriptor.values:
    feature_descriptor_per_feature[f] = cases_descriptor
#word_features_per_pos_tag[pron_pos_tag].append(cases_descriptor.values)
#word_features_per_pos_tag[noun_pos_tag].append(cases_descriptor.values)
#word_features_per_pos_tag[propn_pos_tag].append(cases_descriptor.values)
#word_features_per_pos_tag[adj_pos_tag].append(cases_descriptor.values)
#word_features_per_pos_tag[det_pos_tag].append(cases_descriptor.values)
#word_features_per_pos_tag[num_pos_tag].append(cases_descriptor.values)
#word_features_per_pos_tag[verb_pos_tag].append(cases_descriptor.values)
feature_types_descriptor_per_id[cases_descriptor.id] = cases_descriptor


definiteness_descriptor = FeatureDescriptor('Definite', 'definiteness or state',
                                            'Definiteness is typically a feature of nouns, adjectives and articles. Its value distinguishes whether we are talking about something known and concrete, or something general or unknown.',
                                            'http://universaldependencies.org/u/feat/Definite.html', 'http://universaldependencies.org/es/feat/Definite.html',
                                            ['Definite_Com','Definite_Cons','Definite_Def','Definite_Ind','Definite_Spec'])
for f in definiteness_descriptor.values:
    feature_descriptor_per_feature[f] = definiteness_descriptor
word_features_per_pos_tag[det_pos_tag].append(definiteness_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(definiteness_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(definiteness_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(definiteness_descriptor.values)
feature_types_descriptor_per_id[definiteness_descriptor.id] = definiteness_descriptor


degrees_descriptor = FeatureDescriptor('Degree', 'degree of comparison',
                                       'Degree of comparison is typically an inflectional feature of some adjectives and adverbs.',
                                       'http://universaldependencies.org/u/feat/Degree.html', 'http://universaldependencies.org/es/feat/Degree.html',
                                       ['Degree_Abs','Degree_Cmp','Degree_Equ','Degree_Pos','Degree_Sup'])
for f in degrees_descriptor.values:
    feature_descriptor_per_feature[f] = degrees_descriptor
word_features_per_pos_tag[adj_pos_tag].append(degrees_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(degrees_descriptor.values)
word_features_per_pos_tag[pron_pos_tag].append(degrees_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(degrees_descriptor.values)
feature_types_descriptor_per_id[degrees_descriptor.id] = degrees_descriptor


verb_forms_descriptor = FeatureDescriptor('VerbForm', 'form of verb or deverbative',
                                          'Even though the name of the feature seems to suggest that it is used exclusively with verbs, it is not the case. Some verb forms in some languages actually form a gray zone between verbs and other parts of speech (nouns, adjectives and adverbs).',
                                          'http://universaldependencies.org/u/feat/VerbForm.html', 'http://universaldependencies.org/es/feat/VerbForm.html',
                                          ['VerbForm_Conv','VerbForm_Fin','VerbForm_Gdv','VerbForm_Ger','VerbForm_Inf','VerbForm_Part',
                                           'VerbForm_Sup','VerbForm_Vnoun'])
for f in verb_forms_descriptor.values:
    feature_descriptor_per_feature[f] = verb_forms_descriptor
word_features_per_pos_tag[verb_pos_tag].append(verb_forms_descriptor.values)
word_features_per_pos_tag[aux_pos_tag].append(verb_forms_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(verb_forms_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(verb_forms_descriptor.values)
word_features_per_pos_tag[propn_pos_tag].append(verb_forms_descriptor.values)
word_features_per_pos_tag[x_pos_tag].append(verb_forms_descriptor.values)
word_features_per_pos_tag[sym_pos_tag].append(verb_forms_descriptor.values)
feature_types_descriptor_per_id[verb_forms_descriptor.id] = verb_forms_descriptor


moods_descriptor = FeatureDescriptor('Mood', 'mood',
                                     'Mood is a feature that expresses modality and subclassifies finite verb forms.',
                                     'http://universaldependencies.org/u/feat/Mood.html', 'http://universaldependencies.org/es/feat/Mood.html',
                                     ['Mood_Adm','Mood_Cnd','Mood_Des','Mood_Imp','Mood_Ind','Mood_Jus','Mood_Nec','Mood_Opt','Mood_Pot','Mood_Prp',
                                      'Mood_Qot','Mood_Sub'])
for f in moods_descriptor.values:
    feature_descriptor_per_feature[f] = moods_descriptor
word_features_per_pos_tag[verb_pos_tag].append(moods_descriptor.values)
word_features_per_pos_tag[aux_pos_tag].append(moods_descriptor.values)
feature_types_descriptor_per_id[moods_descriptor.id] = moods_descriptor


tenses_descriptor = FeatureDescriptor('Tense', 'tense',
                                      'Tense is typically a feature of verbs. It may also occur with other parts of speech (nouns, adjectives, adverbs), depending on whether borderline word forms such as gerunds and participles are classified as verbs or as the other category.',
                                      'http://universaldependencies.org/u/feat/Tense.html', 'http://universaldependencies.org/es/feat/Tense.html',
                                      ['Tense_Fut','Tense_Imp','Tense_Past','Tense_Pqp','Tense_Pres'])
for f in tenses_descriptor.values:
    feature_descriptor_per_feature[f] = tenses_descriptor
word_features_per_pos_tag[verb_pos_tag].append(tenses_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(tenses_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(tenses_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(tenses_descriptor.values)
word_features_per_pos_tag[aux_pos_tag].append(tenses_descriptor.values)
feature_types_descriptor_per_id[tenses_descriptor.id] = tenses_descriptor


aspects_descriptor = FeatureDescriptor('Aspect', 'aspect',
                                 'Aspect is typically a feature of verbs. It may also occur with other parts of speech (nouns, adjectives, adverbs), depending on whether borderline word forms such as gerunds and participles are classified as verbs or as the other category.',
                                 'http://universaldependencies.org/u/feat/Aspect.html', 'http://universaldependencies.org/es/feat/Aspect.html',
                                 ['Aspect_Hab','Aspect_Imp','Aspect_Iter','Aspect_Perf','Aspect_Prog','Aspect_Prosp'])
for f in aspects_descriptor.values:
    feature_descriptor_per_feature[f] = aspects_descriptor
word_features_per_pos_tag[verb_pos_tag].append(aspects_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(aspects_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(aspects_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(aspects_descriptor.values)
feature_types_descriptor_per_id[aspects_descriptor.id] = aspects_descriptor


voices_descriptor = FeatureDescriptor('Voice', 'voice',
                                      'Voice is typically a feature of verbs. It may also occur with other parts of speech (nouns, adjectives, adverbs), depending on whether borderline word forms such as gerunds and participles are classified as verbs or as the other category.',
                                      'http://universaldependencies.org/u/feat/Voice.html', 'http://universaldependencies.org/es/feat/Voice.html',
                                      ['Voice_Act','Voice_Antip','Voice_Cau','Voice_Dir','Voice_Inv','Voice_Mid','Voice_Pass','Voice_Rcp'])
for f in voices_descriptor.values:
    feature_descriptor_per_feature[f] = voices_descriptor
word_features_per_pos_tag[verb_pos_tag].append(voices_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(voices_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(voices_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(voices_descriptor.values)
feature_types_descriptor_per_id[voices_descriptor.id] = voices_descriptor


evidentialitys_descriptor = FeatureDescriptor('Evident', 'evidentiality',
                                              'Evidentiality is the morphological marking of a speaker’s source of information (Aikhenvald, 2004). It is sometimes viewed as a category of mood and modality.',
                                              'http://universaldependencies.org/u/feat/Evident.html', 'http://universaldependencies.org/u/feat/Evident.html',
                                              ['Evident_Fh','Evident_Nfh'])
for f in evidentialitys_descriptor.values:
    feature_descriptor_per_feature[f] = evidentialitys_descriptor
feature_types_descriptor_per_id[evidentialitys_descriptor.id] = evidentialitys_descriptor


polaritys_descriptor = FeatureDescriptor('Polarity', 'polarity',
                                         'Polarity is typically a feature of verbs, adjectives, sometimes also adverbs and nouns in languages that negate using bound morphemes. In languages that negate using a function word, Polarity is used to mark that function word, unless it is a pro-form already marked with PronType=Neg (see below).',
                                         'http://universaldependencies.org/u/feat/Polarity.html', 'http://universaldependencies.org/u/feat/Polarity.html',
                                         ['Polarity_Neg','Polarity_Pos'])
for f in polaritys_descriptor.values:
    feature_descriptor_per_feature[f] = polaritys_descriptor
word_features_per_pos_tag[verb_pos_tag].append(polaritys_descriptor.values)
word_features_per_pos_tag[adv_pos_tag].append(polaritys_descriptor.values)
word_features_per_pos_tag[adj_pos_tag].append(polaritys_descriptor.values)
word_features_per_pos_tag[noun_pos_tag].append(polaritys_descriptor.values)
feature_types_descriptor_per_id[polaritys_descriptor.id] = polaritys_descriptor


persons_descriptor = FeatureDescriptor('Person', 'person',
                                       'Person is typically feature of personal and possessive pronouns / determiners, and of verbs. On verbs it is in fact an agreement feature that marks the person of the verb’s subject (some languages, e.g. Basque, can also mark person of objects). Person marked on verbs makes it unnecessary to always add a personal pronoun as subject and thus subjects are sometimes dropped (pro-drop languages).',
                                       'http://universaldependencies.org/u/feat/Person.html', 'http://universaldependencies.org/es/feat/Person.html',
                                       ['Person_0','Person_1','Person_2','Person_3','Person_4'])
for f in persons_descriptor.values:
    feature_descriptor_per_feature[f] = persons_descriptor
word_features_per_pos_tag[pron_pos_tag].append(persons_descriptor.values)
word_features_per_pos_tag[det_pos_tag].append(persons_descriptor.values)
word_features_per_pos_tag[verb_pos_tag].append(persons_descriptor.values)
word_features_per_pos_tag[aux_pos_tag].append(persons_descriptor.values)
word_features_per_pos_tag[x_pos_tag].append(persons_descriptor.values)
word_features_per_pos_tag[sym_pos_tag].append(persons_descriptor.values)
feature_types_descriptor_per_id[persons_descriptor.id] = persons_descriptor


politeness_descriptor = FeatureDescriptor('Polite', 'politeness',
                                 'Various languages have various means to express politeness or respect; some of the means are morphological. Three to four dimensions of politeness are distinguished in linguistic literature. The Polite feature currently covers (and mixes) two of them',
                                 'http://universaldependencies.org/u/feat/Polite.html', 'http://universaldependencies.org/es/feat/Polite.html',
                                 ['Polite_Elev','Polite_Form','Polite_Humb','Polite_Infm'])
word_features_per_pos_tag[pron_pos_tag].append(politeness_descriptor.values)
for f in politeness_descriptor.values:
    feature_descriptor_per_feature[f] = politeness_descriptor
feature_types_descriptor_per_id[politeness_descriptor.id] = politeness_descriptor

for tag in pos_tags:
    word_features_per_pos_tag[tag] = list(itertools.chain.from_iterable(word_features_per_pos_tag[tag]))




from mammut.common.mammut_api import *
from mammut.common.corpus import *
from ipywidgets import *


def to_sp_string(valid: bool):
    if valid:
        return "está correcto."
    else:
        return "presenta errores."


def load_users(simulator_obj, users_filename: str):
    users_file = open(users_filename, "r")
    simulator_obj.users = json.loads(users_file.read())
    users_file.close()


def save_users(simulator_obj, users_filename: str):
    print("users", simulator_obj.users)
    users_file = open(users_filename, "w")
    users_file.write(json.dumps(simulator_obj.users))
    users_file.close()


def add_new_user(simulator_obj, user, user_filename, type):
    print("user sssss", simulator_obj.users)
    if type not in simulator_obj.users.keys():
        simulator_obj.users[type] = dict()
    if user.first_name is not "":
        simulator_obj.users[type][user.first_name] = int(user.id)
    else:
        simulator_obj.users[type][user.main_email] = int(user.id)
    save_users(simulator_obj, user_filename)


def init_users(simulator_obj, users_filename: str):
    # Init users
    print("init user", simulator_obj.users)
    if not os.path.isfile(users_filename):
        simulator_obj.current_user = simulator_obj.mammut_api.get_fixed_mammut()
        add_new_user(
            simulator_obj,
            simulator_obj.current_user,
            users_filename,
            UserType.machine.value,
        )
    else:
        load_users(simulator_obj, users_filename)


def load_rooms(simulator_obj, rooms_filename: str):
    room_file = open(rooms_filename, "r")
    simulator_obj.rooms = json.loads(room_file.read())
    room_file.close()


def save_rooms(simulator_obj, rooms_filename: str):
    rooms_file = open(rooms_filename, "w")
    print("rooms", simulator_obj.rooms)
    rooms_file.write(json.dumps(simulator_obj.rooms))
    rooms_file.close()


def add_new_room(simulator_obj, room, rooms_filename, type, extra_suffix: str = None):
    if type not in simulator_obj.rooms.keys():
        simulator_obj.rooms[type] = dict()
    if extra_suffix is None:
        simulator_obj.rooms[type][room.title] = int(room.id)
    else:
        simulator_obj.rooms[type][room.title + extra_suffix] = int(room.id)
    save_rooms(simulator_obj, rooms_filename)


def init_rooms(simulator_obj, rooms_filename: str):
    # Init users
    print("rooms load")
    if os.path.isfile(rooms_filename):
        print("rooms load")
        load_rooms(simulator_obj, rooms_filename)
        print("rooms in simulator", simulator_obj.rooms)



# coding=utf-8

from collections import *
from nltk.tokenize.regexp import RegexpTokenizer, re
from nltk.probability import FreqDist
from enum import Enum
import sys

from typing import List, Tuple
from mammut.common import WORD_LEXICAL_RELEVANCE_INDEX_ESTIMATOR_ADD_NEW_SAMPLE
from mammut.models.lexical_relevance import LexicalRelevanceIndexEstimator

class PuntuationType(Enum):
    Dot = 1
    Comma = 2
    Exclamation = 3
    Interrogation = 4
    Column = 5
    SemiColumn = 6


class BasicTokenizerResult:
    # expresion regular para espacios en blanco en espanol
    WHITE_SPACE_REGEX_CHARS = r" \t"
    WHITE_SPACE_REGEX_STR = "[" + WHITE_SPACE_REGEX_CHARS + "]*"
    WHITE_SPACE_REGEX = re.compile(WHITE_SPACE_REGEX_STR)
    # expresion regular para palabras en espanol
    WORD_REGEX_CHARS = r"A-ZÑÁÉÍÓÚÜÇa-zñáéíóúüç\-"
    WORD_REGEX_STR = "[" + WORD_REGEX_CHARS + "]+"
    WORD_REGEX = re.compile(WORD_REGEX_STR)
    # expresion regular para numeros
    NUMBER_REGEX_CHARS = r"0-9"
    NUMBER_REGEX_STR = "[" + NUMBER_REGEX_CHARS + "]+"
    NUMBER_REGEX = re.compile(NUMBER_REGEX_STR)
    # expresion regular para signos de puntuacion
    PUNCTUATION_REGEX_CHARS = r"\.,!¡\?¿:;"
    PUNCTUATION_REGEX_STR = (
        r"\.{3}|[" + PUNCTUATION_REGEX_CHARS + "]" + WHITE_SPACE_REGEX_STR
    )
    PUNCTUATION_REGEX = re.compile(PUNCTUATION_REGEX_STR)
    # expresion regular para comillas
    QUOTE_REGEX_CHARS = r'"\''
    QUOTE_REGEX_STR = "[" + QUOTE_REGEX_CHARS + "]" + WHITE_SPACE_REGEX_STR
    QUOTE_REGEX = re.compile(QUOTE_REGEX_STR)
    # expresion regular para comillas
    PARENTHESIS_REGEX_CHARS = r"\(\)"
    PARENTHESIS_REGEX_STR = "[" + PARENTHESIS_REGEX_CHARS + "]" + WHITE_SPACE_REGEX_STR
    PARENTHESIS_REGEX = re.compile(PARENTHESIS_REGEX_STR)
    # expresion regular para finales de linea
    BREAK_REGEX_CHARS = r"\n"
    BREAK_REGEX_STR = WHITE_SPACE_REGEX_STR + BREAK_REGEX_CHARS
    BREAK_REGEX = re.compile(BREAK_REGEX_STR)
    # expresion regular para encontrar variables
    VARIABLE_REGEX_STR = r"\[[Vv]ariable\|[^\]]+\]"
    VARIABLE_REGEX = re.compile(VARIABLE_REGEX_STR)
    # expresion regular para encontrar images
    IMAGE_REGEX_STR = r"\[[Ii]mage\|[^\]]+\]"
    IMAGE_REGEX = re.compile(IMAGE_REGEX_STR)
    # expresion regular para encontrar botones
    VERBS_REGEX_STR = r"\{\{[Kk][Bb]\-[0-9]+\}\}"
    VERBS_REGEX = re.compile(VERBS_REGEX_STR)
    # expression regular para elementos desconocidos
    UNKNOWN_REGEX_STR = r"[^"
    UNKNOWN_REGEX_STR += WORD_REGEX_CHARS
    UNKNOWN_REGEX_STR += NUMBER_REGEX_CHARS
    UNKNOWN_REGEX_STR += PUNCTUATION_REGEX_CHARS
    UNKNOWN_REGEX_STR += QUOTE_REGEX_CHARS
    UNKNOWN_REGEX_STR += PARENTHESIS_REGEX_CHARS
    UNKNOWN_REGEX_STR += r"\s]+"

    def __init__(
        self,
        base_result: List[str],
        original_text: str,
        stats,
        layer: int,
        regional_settings: str,
        knowledge_graph=None,
    ):
        self.base_result = base_result
        self.original_text = original_text
        self.knowledge_graph = knowledge_graph
        self.regional_settings = regional_settings
        self.words = []
        self.numbers = []
        self.puntuations = []
        self.quotes = []
        self.parenthesis = []
        self.breaks = []
        self.variables = []
        self.invalid_variables = []
        self.images = []
        self.invalid_images = []
        self.verbs = []
        self.invalid_verbs = []
        self.unknowns = []
        self.token_offsets = []
        self.sentence_offsets = []
        self._token_cursor = 0
        self._sentence_cursor = 0
        offset_tuple = None
        for item in self.base_result:
            # Handle separators
            separator = ""
            local_token_cursor = self._token_cursor
            while self.original_text[local_token_cursor] != item[0]:
                separator += self.original_text[local_token_cursor]
                local_token_cursor += 1
            if separator:
                offset_tuple = self.create_token_offset(separator)
                if "\n" in separator:
                    self.create_sentence_offset(offset_tuple)
            # Handle tokens
            if self.is_word(item):
                if stats:
                    word = stats.add_word(item, layer)
                else:
                    word = item.lower()
                self.words.append(word)
                offset_tuple = self.create_token_offset(item)
            elif self.is_number(item):
                if stats:
                    stats.add_number(item)
                self.numbers.append(item)
                offset_tuple = self.create_token_offset(item)
            elif self.is_punctuation(item):
                if stats:
                    stats.add_puntuation(item)
                self.puntuations.append(item)
                offset_tuple = self.create_token_offset(item)
            elif self.is_quote(item):
                if stats:
                    stats.add_quote(item)
                self.quotes.append(item)
                offset_tuple = self.create_token_offset(item)
            elif self.is_parenthesis(item):
                if stats:
                    stats.add_parenthesis(item)
                self.parenthesis.append(item)
                offset_tuple = self.create_token_offset(item)
            elif self.is_break(item):
                if stats:
                    stats.add_break(item)
                self.breaks.append(item)
                offset_tuple = self.create_token_offset(item)
            elif self.is_variable(item):
                variable_name = self.get_variable_name(item)
                if stats:
                    stats.add_variable(variable_name)
                self.variables.append(variable_name)
                if (
                    self.knowledge_graph is not None
                    and not self.knowledge_graph.is_valid_variable(variable_name)
                ):
                    self.invalid_variables.append(variable_name)
                offset_tuple = self.create_token_offset(item)
            elif self.is_image(item):
                image_name = self.get_image_name(item)
                if stats:
                    stats.add_image(image_name)
                self.images.append(image_name)
                if (
                    self.knowledge_graph is not None
                    and not self.knowledge_graph.is_valid_image(image_name)
                ):
                    self.invalid_images.append(image_name)
                offset_tuple = self.create_token_offset(item)
            elif self.is_verb(item):
                verb_id = self.get_verb_id(item)
                if stats:
                    stats.add_verb(verb_id)
                self.verbs.append(verb_id)
                if (
                    self.knowledge_graph is not None
                    and not self.knowledge_graph.is_valid_verb(verb_id)
                ):
                    self.invalid_verbs.append(verb_id)
                offset_tuple = self.create_token_offset(item)
            else:
                if stats:
                    unknown = stats.add_unknown(item)
                else:
                    unknown = item.lower()
                self.unknowns.append(unknown)
                offset_tuple = self.create_token_offset(item)
        # Handle last separators
        separator = ""
        local_token_cursor = self._token_cursor
        while local_token_cursor < len(self.original_text):
            separator += self.original_text[local_token_cursor]
            local_token_cursor += 1
        if separator:
            offset_tuple = self.create_token_offset(separator)
        if offset_tuple:
            self.create_sentence_offset(offset_tuple)

    def create_token_offset(self, token: str):
        offset_tuple = (self._token_cursor, self._token_cursor + (len(token) - 1))
        self._token_cursor = offset_tuple[1] + 1
        self.token_offsets.append(offset_tuple)
        return offset_tuple

    def create_sentence_offset(self, token_offset_tuple: Tuple):
        offset_tuple = (self._sentence_cursor, token_offset_tuple[1])
        self._sentence_cursor = offset_tuple[1] + 1
        self.sentence_offsets.append(offset_tuple)
        return offset_tuple

    def get_variable_name(self, item: str):
        variable_name = item.lower().strip("[]")
        variable_name = variable_name.replace("variable|", "")
        return variable_name

    def get_image_name(self, item: str):
        image_name = item.lower().strip("[]")
        image_name = image_name.replace("image|", "")
        return image_name

    def get_verb_id(self, item: str):
        verb_id = item.upper().strip("{}")
        return verb_id

    def is_word(self, item: str):
        res = BasicTokenizerResult.WORD_REGEX.match(item)
        return res != None

    def is_number(self, item: str):
        res = BasicTokenizerResult.NUMBER_REGEX.match(item)
        return res != None

    def is_punctuation(self, item: str):
        res = BasicTokenizerResult.PUNCTUATION_REGEX.match(item)
        return res != None

    def is_quote(self, item: str):
        res = BasicTokenizerResult.QUOTE_REGEX.match(item)
        return res != None

    def is_parenthesis(self, item: str):
        res = BasicTokenizerResult.PARENTHESIS_REGEX.match(item)
        return res != None

    def is_break(self, item: str):
        res = BasicTokenizerResult.BREAK_REGEX.match(item)
        return res != None

    def is_variable(self, item: str):
        res = BasicTokenizerResult.VARIABLE_REGEX.match(item)
        return res != None

    def is_image(self, item: str):
        res = BasicTokenizerResult.IMAGE_REGEX.match(item)
        return res != None

    def is_verb(self, item: str):
        res = BasicTokenizerResult.VERBS_REGEX.match(item)
        return res != None

    def contains_unknown(self):
        return len(self.unknowns) > 0

    def contains_invalid_variables(self):
        return len(self.invalid_variables) > 0

    def contains_invalid_images(self):
        return len(self.invalid_images) > 0

    def contains_invalid_verbs(self):
        return len(self.invalid_verbs) > 0

    def print(self):
        print("Resultado Base: \n{0}".format(self.base_result))
        print("Palabras: \n{0}".format(self.words))
        print("Numeros: \n{0}".format(self.numbers))
        print("Signos de puntuacion: \n{0}".format(self.puntuations))
        print("Comillas: \n{0}".format(self.quotes))
        print("Parentesis: \n{0}".format(self.parenthesis))
        print("Variables: \n{0}".format(self.variables))
        print("Botones: \n{0}".format(self.verbs))
        print("Desconocidas: \n{0}".format(self.unknowns))


class BasicTokenizerStatistics:
    BasicTokenizerStatisticsSnapshot = namedtuple(
        "BasicTokenizerStatisticsSnapshot",
        ["words_count", "undefined_morphemes_count", "definition"],
    )

    def __init__(self, regional_settings: str):
        self.regional_settings = regional_settings
        self.words = FreqDist()
        self.word_list = []
        self.word_layer = {}
        self.word_lexical_relevance_index_estimator = LexicalRelevanceIndexEstimator(
            self.regional_settings, self.words
        )
        self.numbers = FreqDist()
        self.puntuations = FreqDist()
        self.quotes = FreqDist()
        self.parenthesis = FreqDist()
        self.breaks = FreqDist()
        self.variables = FreqDist()
        self.images = FreqDist()
        self.verbs = FreqDist()
        self.unknowns = FreqDist()
        self.definitions = 0
        self.corpus_entries = 0
        self.snapshots = []

    def add_word(self, word: str, layer: int):
        word_lower = word.lower()
        if self.words[word_lower] == 0:
            self.word_list.append(word_lower)
        if word_lower not in self.word_layer or self.word_layer[word_lower] > layer:
            self.word_layer[word_lower] = layer
        if WORD_LEXICAL_RELEVANCE_INDEX_ESTIMATOR_ADD_NEW_SAMPLE:
            self.word_lexical_relevance_index_estimator.add_new_sample(word_lower)
        self.words[word_lower] += 1
        return word_lower

    def add_number(self, number: str):
        self.numbers[number] += 1

    def add_puntuation(self, puntuation: str):
        self.puntuations[puntuation] += 1

    def add_quote(self, quote: str):
        self.quotes[quote] += 1

    def add_parenthesis(self, parenth: str):
        self.parenthesis[parenth] += 1

    def add_break(self, parenth: str):
        self.breaks[parenth] += 1

    def add_variable(self, variable_name: str):
        self.variables[variable_name] += 1
        return variable_name

    def add_image(self, image_name: str):
        self.images[image_name] += 1
        return image_name

    def add_verb(self, verb_id: str):
        self.verbs[verb_id] += 1
        return verb_id

    def add_unknown(self, unknown: str):
        unknown_lower = unknown.lower()
        self.unknowns[unknown_lower] += 1
        return unknown_lower

    def get_words_generator(self, layer: int = 0, with_index: bool = False):
        current_index = 0
        while current_index < len(self.word_list):
            w = self.word_list[current_index]
            if self.word_layer[w] >= 0 and self.word_layer[w] <= layer:
                if with_index:
                    lexical_relevance_index = self.word_lexical_relevance_index_estimator.get_index(
                        w
                    )
                    yield w, lexical_relevance_index
                else:
                    yield w
            current_index += 1

    def get_word_lexical_relevance_index(self, sample, add):
        return self.word_lexical_relevance_index_estimator.get_index(sample, add)

    def is_word_under_layer(self, word: str, layer=0):
        return word in self.get_words_generator(layer)

    def get_words_count(self):
        count = 0
        current_index = 0
        while current_index < len(self.word_list):
            w = self.word_list[current_index]
            if self.word_layer[w] >= 0:
                count = count + 1
            current_index += 1
        return count

    def create_snapshot(self, undefined_morphemes_count, definition):
        snapshot = BasicTokenizerStatistics.BasicTokenizerStatisticsSnapshot(
            self.get_words_count(), undefined_morphemes_count, definition
        )
        self.snapshots.append(snapshot)

    def get_plot_data(self):
        plot_data_words_count = []
        plot_data_words_count_colors = []
        plot_data_undefined_morphemes_count = []
        plot_data_undefined_morphemes_count_colors = []
        for ss in self.snapshots:
            plot_data_words_count.append(ss.words_count)
            plot_data_undefined_morphemes_count.append(ss.undefined_morphemes_count)
            if ss.definition:
                plot_data_words_count_colors.append(0)
                plot_data_undefined_morphemes_count_colors.append(0)
            else:
                plot_data_words_count_colors.append(1)
                plot_data_undefined_morphemes_count_colors.append(1)
        return (
            plot_data_words_count,
            plot_data_words_count_colors,
            plot_data_undefined_morphemes_count,
            plot_data_undefined_morphemes_count_colors,
        )

    def print(self):
        print("Estadisticas:")
        print("Palabras: \n{0}".format(self.words.items()))
        print("Numeros: \n{0}".format(self.numbers.items()))
        print("Signos de puntuacion: \n{0}".format(self.puntuations.items()))
        print("Comillas: \n{0}".format(self.quotes.items()))
        print("Parentesis: \n{0}".format(self.parenthesis.items()))
        print("Variables: \n{0}".format(self.variables.items()))
        print("Botones: \n{0}".format(self.verbs.items()))
        print("Desconocidas: \n{0}".format(self.unknowns.items()))


class BasicTokenizerStatisticsDict(defaultdict):
    def __init__(self):
        super(BasicTokenizerStatisticsDict, self).__init__(BasicTokenizerStatistics)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class BasicTokenizer(RegexpTokenizer):
    """http://www.nltk.org/_modules/nltk/tokenize/regexp.html"""

    def __init__(
        self,
        gaps=False,
        discard_empty=True,
        flags=re.UNICODE | re.MULTILINE | re.DOTALL,
    ):
        self.dictionary = None
        self.stats = BasicTokenizerStatisticsDict()
        pattern = BasicTokenizerResult.VARIABLE_REGEX_STR
        pattern += "|" + BasicTokenizerResult.IMAGE_REGEX_STR
        pattern += "|" + BasicTokenizerResult.VERBS_REGEX_STR
        pattern += "|" + BasicTokenizerResult.WORD_REGEX_STR
        pattern += "|" + BasicTokenizerResult.NUMBER_REGEX_STR
        pattern += "|" + BasicTokenizerResult.PUNCTUATION_REGEX_STR
        pattern += "|" + BasicTokenizerResult.QUOTE_REGEX_STR
        pattern += "|" + BasicTokenizerResult.PARENTHESIS_REGEX_STR
        pattern += "|" + BasicTokenizerResult.BREAK_REGEX_STR
        pattern += "|" + BasicTokenizerResult.UNKNOWN_REGEX_STR
        RegexpTokenizer.__init__(self, pattern, gaps, discard_empty, flags)

    def set_dictionary(self, dictionary):
        self.dictionary = dictionary

    def _tokenize(
        self,
        text: str,
        definition: bool,
        layer: int,
        regional_settings: str,
        knowledge_graph=None,
        exclude_from_stats: bool = False,
    ) -> BasicTokenizerResult:
        base_result = RegexpTokenizer.tokenize(self, text)
        stats = None
        if not exclude_from_stats:
            stats = self.stats[regional_settings]
        result = BasicTokenizerResult(
            base_result, text, stats, layer, regional_settings, knowledge_graph
        )
        undefined_morphemes_count = 0
        if self.dictionary is not None:
            undefined_morphemes_count = self.dictionary.get_undefined_morphemes_count(
                regional_settings
            )
        if not exclude_from_stats:
            self.stats[regional_settings].create_snapshot(
                undefined_morphemes_count, definition
            )
        return result

    def tokenize_definition(self, text: str, layer: int, regional_settings: str, knowledge_graph):
        result = self._tokenize(text, True, layer, regional_settings, knowledge_graph)
        self.stats[regional_settings].definitions += 1
        return result

    def tokenize_corpus_entry(self, text: str, add_to_layer_zero: bool, regional_settings: str):
        # the -1 layer is used to indicate that any word in the text must not be
        # counted as word to define.
        layer = -1
        if add_to_layer_zero:
            layer = 0
        result = self._tokenize(text, False, layer, regional_settings)
        self.stats[regional_settings].corpus_entries += 1
        return result

    def tokenize_out_of_stats(self, text: str, regional_settings: str, knowledge_graph=None):
        result = self._tokenize(
            text, True, sys.maxsize, regional_settings, knowledge_graph, True
        )
        return result



from threading import Timer

import mammut
from mammut.common.mammut_api import *
from mammut.common.simulator_utilities import *
from ipywidgets import *
from datetime import datetime, date
import logging
from django.template.loader import get_template
import mammut.visualization.user_form as user_form
import mammut.visualization.room_form as room_form
import mammut.visualization.solve_thought_offer_to_be_learn as solve_thought_offer_to_be_learn
from mammut.visualization.simulator_widget_operation_utils import get_operations_boxes
from mammut.visualization.simulator_widget_simulation_utils import get_simulation_boxes
from mammut.visualization.simulator_widget_compilation_utils import (
    get_compilation_boxes,
)
import enum


log = logging.getLogger(__name__)

# quien debe instanciar el simulator_obj??
def get_simulator_widget(
    simulator_obj,
    mammut_api_url,
    redis_address,
    redis_port,
    package_file,
    standard_spreadsheet_id,
    users_filename,
    users_rooms_spreadsheet_id,
    users_sheet_name,
    rooms_sheet_name,
    rooms_filename,
):
    def move_to_next_box():
        accordion_size = len(accordion.children)
        if accordion.selected_index < accordion_size:
            accordion.selected_index = accordion.selected_index + 1
        else:
            accordion.selected_index = 0

    ## Select API URL ##
    def start_button_clicked(b):
        # clear_output()
        simulator_obj.mammut_api = Request(apiUrl_textBox.value)
        init_users(simulator_obj, users_filename)
        load_rooms(simulator_obj, rooms_filename)
        # Updates Alex Mammut #
        alex = simulator_obj.mammut_api.get_fixed_mammut()
        add_new_user(simulator_obj, alex, users_filename, UserType.machine.value)
        move_to_next_box()

    apiUrl_textBox = widgets.Text(
        value=mammut_api_url,
        description="Mammut API Url:",
        layout=Layout(width="100%", height="50px"),
    )
    redisUrl_textBox = widgets.Text(
        value=redis_address,
        description="Redis Url:",
        layout=Layout(width="100%", height="50px"),
    )
    start_button = widgets.Button(description="Comenzar", button_style="primary")
    configurationBox = widgets.VBox(
        [widgets.HBox([apiUrl_textBox]), redisUrl_textBox, start_button]
    )
    start_button.on_click(start_button_clicked)

    ## Select your operation mode ##
    class operation_modes(enum.Enum):
        COMPILATION = "Compilation"
        SIMULATION = "Simulation"
        OPERATIONS = "Operations"

    operation_modes_dict = {
        operation_modes.COMPILATION.value: operation_modes.COMPILATION.value,
        operation_modes.SIMULATION.value: operation_modes.SIMULATION.value,
        operation_modes.OPERATIONS.value: operation_modes.OPERATIONS.value,
    }
    operation_modes_dropdown = widgets.Dropdown(
        options=operation_modes_dict,
        value=operation_modes_dict[operation_modes.COMPILATION.value],
        description="Select your notebook operation mode",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    select_operation_mode_button = widgets.Button(
        description="Select", button_style="primary", style={"description_width": "50%"}
    )
    operation_modes_box = widgets.VBox(
        [widgets.HBox([operation_modes_dropdown]), select_operation_mode_button]
    )
    operation_mode_state = None

    def select_operation_mode(b):
        global operation_mode_state
        operation_mode_state = operation_modes(operation_modes_dropdown.value)
        if operation_mode_state == operation_modes.COMPILATION:
            new_boxes = get_compilation_boxes(
                simulator_obj, package_file, standard_spreadsheet_id, move_to_next_box
            )
            accordion.children = accordion.children + new_boxes
            accordion.set_title(2, "2. Seleccionar bot")
            accordion.set_title(3, "3. Validar Archivo")
            accordion.set_title(4, "4. Actualizar simulador")
            accordion.set_title(5, "5. Image retraining")

            def get_image_retrain_widget(b):
                if simulator_obj.package is not None:
                    if accordion.selected_index == 5:
                        image_retraining_widget = simulator_obj.package.get_image_retraining_widget(
                            simulator_obj.current_user
                        )
                        curr_childrens = list(accordion.children)
                        image_box = curr_childrens[5]
                        image_box.children = [image_retraining_widget]

            accordion.observe(get_image_retrain_widget)

        elif operation_mode_state == operation_modes.SIMULATION:
            new_boxes = get_simulation_boxes(simulator_obj, move_to_next_box)
            accordion.children = accordion.children + new_boxes
            accordion.set_title(2, "2. Seleccionar bot")
            accordion.set_title(3, "3. Seleccionar Usuario")
            accordion.set_title(4, "4. Seleccionar Room")
            accordion.set_title(5, "5. Enviar mensaje")
            accordion.set_title(6, "6. Solve Issue")

        elif operation_mode_state == operation_modes.OPERATIONS:
            new_boxes = get_operations_boxes(
                simulator_obj,
                users_rooms_spreadsheet_id,
                users_sheet_name,
                rooms_sheet_name,
                apiUrl_textBox.value,
                users_filename,
                rooms_filename,
                move_to_next_box,
            )
            accordion.children = accordion.children + new_boxes
            accordion.set_title(3, "Crear nuevas entidades")
            accordion.set_title(2, "Seleccionar bot")
            accordion.set_title(4, "Cargar Cargar Usuarios y Chats")
            accordion.set_title(5, "Cargar video")
            accordion.set_title(6, "Cargar Calendario")

        else:
            raise Exception("Unknown simulator operation mode.")

        move_to_next_box()

    select_operation_mode_button.on_click(select_operation_mode)
    ## /Select your operation mode ##
    ### Main Accordion ###
    accordion = widgets.Accordion(children=[configurationBox, operation_modes_box])
    accordion.set_title(0, "0. Configuracion")
    accordion.set_title(1, "1. Seleccionar operation mode")
    accordion.selected_index = 0
    return accordion



# coding=utf-8

import tensorflow as tf
import numpy as np

from mammut.models.parameter_node import ParameterNode
from mammut.models.worker_node import WorkerNode
from mammut.common.util import StringEnum, RegionalSettingsType


class ExecutionMode(StringEnum):
    eager = "eager"
    graph = "graph"
    differentiable = "differentiable"


class TransitionOracleMockNode(WorkerNode):
    def __init__(self):
        # TODO:cambiar nombres de parametro que se pasan a constructor de workernode
        WorkerNode.__init__(self, "worker-mx", "model-test", "worker-mx", 0, None)
        self.counter = 0
        # se sobreescribre para cada caso en las pruebas
        # transitoins debe ser un tensor
        self.transition_list = [
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0],
            ],
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0],
            ],
        ]
        self.transitions = tf.constant(
            self.transition_list, dtype=tf.float32
        )  # shape = steps, batch, numero de transiciones posibles

    def _operation(self, *args, **kwargs):
        if len(args[0]) > 1:
            if isinstance(args[0][1], tf.Tensor):
                # print('arg0', args[0][1], isinstance(args[0][1], tf.Tensor))
                transition = self.transitions[args[0][1]]
        else:
            if self.transitions.shape[0] > self.counter:
                transition = self.transitions[self.counter]
                self.counter += 1
            else:
                transition = None  # [[0,0,0,0,0,0,0,0]]
            return transition
        return transition


class OutputOracleMockNode(WorkerNode):
    def __init__(self):
        # TODO:cambiar nombres de parametro que se pasan a constructor de workernode
        WorkerNode.__init__(self, "worker-mx", "model-test", "worker-mx", 0, None)
        self.counter = 0
        # debe ser un tensor!!???
        self.outputs = [
            [[1], [2], [3]],
            [[1], [2], [3]],
        ]  # shape = steps, batch, output_value(numero de outputs posibles si fuese onehot)

    def _operation(self, *args, **kwargs):
        print("arg0", args[0])
        print("arg1", args[0][1], isinstance(args[0][1], tf.Tensor))
        if len(args[0]) > 1:
            if isinstance(args[0][1], tf.Tensor):
                print("arg0", args[0][1], isinstance(args[0][1], tf.Tensor))
                outputs = self.outputs[args[0][1]]
                self.counter += 1
            else:
                outputs = self.outputs[args[0][1]]
                self.counter += 1
        else:
            if self.outputs.shape[0] > self.counter:
                outputs = self.outputs[self.counter]
                self.counter += 1
            else:
                outputs = None  # [[0,0,0,0,0,0,0,0]]
        return outputs


class TransitionSystemConfigurations(ParameterNode):
    def __init__(
        self,
        stack_size: int,
        transition_size: int,
        output_size: int,
        input_size: int,
        batch_size: int,
        stack_position: int,
        stack_position_elems: list,
        transition_position: int,
        output_position: list,
        input_position: int,
        execution_mode: ExecutionMode,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        redis_connection_pool,
    ):
        ParameterNode.__init__(
            self,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )
        self.stack_size = stack_size  # deberia buscarse bien cual es la longitud optima, longitud de la palabra mas largs
        self.transition_size = transition_size
        self.output_size = output_size  # ids de entities, lemmas, etc, #numero de states pasados que sirven de contexto
        self.input_size = input_size
        self.batch_size = batch_size
        self.state_size = (
            self.stack_size + self.transition_size + self.output_size + self.input_size
        )
        self.stack_position = stack_position  # 0-4
        self.stack_position_elems = stack_position_elems
        self.transition_position = transition_position
        self.output_position = output_position
        self.input_position = input_position
        self.transitions = None  # transiciones a aplicar
        self.transitions_array = None
        self.step_transition = None
        # TODO: los names deberian estr hardcode???
        # TODO: en sentido estricto add_output y add_transition no son trnasiciones, se calculan al final y se suman al state resultante
        self.transition_names = [
            "shift",
            "discard",
            "delete_stack",
            "output_shift",
            "morpheme",
        ]
        self.available_transitions_len = len(
            self.transition_names
        )  # numero de transiciones posibles
        self.available_transitions = (
            self.get_one_hot_transitions()
        )  # transiciones que s epueden ejecutar
        self.execution_mode = execution_mode

    def _operation(self, *args, **kwargs):
        return

    ######### SHIFT######

    @staticmethod
    def get_selector_shift_projection(transitions, transition_unit_vector):
        transitions_number = len(transition_unit_vector)
        selector_shift_projection = tf.constant(
            [transition_unit_vector], dtype=tf.float32, shape=[transitions_number, 1]
        )  # numero de transiciones en dimension 1
        selector_shift_projection_result = tf.matmul(
            transitions, selector_shift_projection
        )
        return selector_shift_projection_result

    def get_selector_shift_filter(self, state_size):
        tensor_ones_indexes = [k for k in range(0, state_size)]
        tensor_ones_indexes = np.roll(tensor_ones_indexes, 1)

        selector_shift_filter = np.zeros([state_size, state_size])
        for i in range(0, state_size):
            if (
                i != self.transition_position
                and i not in self.output_position
                and i != self.input_position
            ):
                selector_shift_filter[i, tensor_ones_indexes[i]] = 1
            if (
                i in self.output_position
            ):  # or i == transition_position: #mantener el transtion??o conviene eliminarla de una vez
                selector_shift_filter[i, i] = 1
            if i == self.transition_position:
                selector_shift_filter[i, i] = 0
        return selector_shift_filter

    def get_selector_shift_result(
        self, transitions, transition_unit_vector, state_size
    ):
        selector_shift_projection_result = self.get_selector_shift_projection(
            transitions, transition_unit_vector
        )  # segundo es lo puedo pasar como transtions[index]
        selector_shift_filter = self.get_selector_shift_filter(state_size)

        selector_shift_filter = tf.reshape(
            tf.constant(selector_shift_filter, dtype=tf.float32),
            shape=[1, state_size, state_size],
        )
        selector_shift_projection_result_reshape = tf.reshape(
            selector_shift_projection_result,
            shape=[selector_shift_projection_result.shape[0], 1, 1],
        )
        selector_shift_result = (
            selector_shift_projection_result_reshape * selector_shift_filter
        )
        return selector_shift_result

    ##########DISCARD################

    def get_selector_clean_input_projection(self, transitions, transition_unit_vector):
        selector_clean_input_projection = tf.constant(
            [transition_unit_vector],
            dtype=tf.float32,
            shape=[self.available_transitions_len, 1],
        )
        selector_clean_input_projection_result = tf.matmul(
            transitions, selector_clean_input_projection
        )
        return selector_clean_input_projection_result

    def get_selector_clean_input_filter(self, state_size, input_position):
        selector_clean_input_filter = np.zeros([state_size, state_size])
        for i in range(0, state_size):
            if i != input_position:
                selector_clean_input_filter[i, i] = 1
            else:
                selector_clean_input_filter[i, i] = 0
            if i == self.transition_position:
                selector_clean_input_filter[i, i] = 0
        return selector_clean_input_filter

    def get_selector_clean_input_result(
        self, transitions, transition_unit_vector, state_size
    ):

        selector_clean_input_filter = self.get_selector_clean_input_filter(
            state_size, self.input_position
        )
        selector_clean_input_filter = tf.reshape(
            tf.constant(selector_clean_input_filter, dtype=tf.float32),
            shape=[1, state_size, state_size],
        )
        # print("clean_input Filter: \n %s" % (selector_clean_input_filter.eval()))
        selector_clean_input_projection_result = self.get_selector_clean_input_projection(
            transitions, transition_unit_vector
        )

        selector_clean_input_projection_result_reshape = tf.reshape(
            selector_clean_input_projection_result,
            shape=[selector_clean_input_projection_result.shape[0], 1, 1],
        )
        # print("clean_input transitions reshaped: \n %s" % (selector_clean_input_projection_result_reshape.eval()))

        selector_clean_input_projection_result = (
            selector_clean_input_projection_result_reshape * selector_clean_input_filter
        )
        # print("clean input transitions: \n %s" % selector_clean_input_projection_result.eval())
        return selector_clean_input_projection_result

    ################DELETE STACK##################

    def get_selector_delete_stack_projection(self, transitions, transition_unit_vector):
        selector_delete_stack_projection = tf.constant(
            [transition_unit_vector],
            dtype=tf.float32,
            shape=[self.available_transitions_len, 1],
        )
        selector_delete_stack_projection_result = tf.matmul(
            transitions, selector_delete_stack_projection
        )
        return selector_delete_stack_projection_result

    def get_selector_delete_stack_filter(self, state_size, stack_postion_elems):
        selector_delete_stack_filter = np.zeros([state_size, state_size])
        for i in range(0, state_size):
            if i in stack_postion_elems:
                selector_delete_stack_filter[i, i] = 0
            else:
                selector_delete_stack_filter[i, i] = 1
            if i == self.transition_position:
                selector_delete_stack_filter[i, i] = 0
        return selector_delete_stack_filter

    def get_selector_delete_stack_result(
        self, transitions, transition_unit_vector, state_size
    ):

        selector_delete_stack_filter = self.get_selector_delete_stack_filter(
            state_size, self.stack_position_elems
        )
        selector_delete_stack_filter = tf.reshape(
            tf.constant(selector_delete_stack_filter, dtype=tf.float32),
            shape=[1, state_size, state_size],
        )
        # print("delete_stack Filter: \n %s" % (selector_delete_stack_filter.eval()))

        selector_delete_stack_projection_result = self.get_selector_delete_stack_projection(
            transitions, transition_unit_vector
        )
        selector_delete_stack_projection_result_reshape = tf.reshape(
            selector_delete_stack_projection_result,
            shape=[selector_delete_stack_projection_result.shape[0], 1, 1],
        )
        # print("delete_stack transitions reshaped: \n %s" % (selector_delete_stack_projection_result_reshape.eval()))

        selector_delete_stack_projection_result = (
            selector_delete_stack_projection_result_reshape
            * selector_delete_stack_filter
        )
        # print("delete stack transitions: \n %s" % selector_delete_stack_projection_result.eval())
        return selector_delete_stack_projection_result

    ################## OUTPUT SHIFT ###################

    def get_selector_shift_output_projection(self, transitions, transition_unit_vector):
        selector_shift_output_projection = tf.constant(
            [transition_unit_vector],
            dtype=tf.float32,
            shape=[self.available_transitions_len, 1],
        )
        selector_shift_output_projection_result = tf.matmul(
            transitions, selector_shift_output_projection
        )
        return selector_shift_output_projection_result

    def get_selector_shift_output_filter(self, state_size, morpheme_position):
        tensor_ones_indexes = [k for k in range(0, state_size)]
        tensor_ones_indexes = np.roll(tensor_ones_indexes, 1)

        selector_shift_output_filter = np.zeros([state_size, state_size])
        for i in range(0, state_size):
            if i in morpheme_position[1:]:
                selector_shift_output_filter[i, tensor_ones_indexes[i]] = 1
            else:
                selector_shift_output_filter[i, i] = 0
            if i == self.transition_position:
                selector_shift_output_filter[i, i] = 0
        return selector_shift_output_filter

    def get_selector_shift_output_result(
        self, transitions, transition_unit_vector, state_size
    ):
        selector_shift_output_filter = self.get_selector_shift_output_filter(
            state_size, self.output_position
        )
        selector_shift_output_filter = tf.reshape(
            tf.constant(selector_shift_output_filter, dtype=tf.float32),
            shape=[1, state_size, state_size],
        )
        # print("shift_output Filter: \n %s" % (selector_shift_output_filter.eval()))

        selector_shift_output_projection_result = self.get_selector_shift_output_projection(
            transitions, transition_unit_vector
        )
        selector_shift_output_projection_result_reshape = tf.reshape(
            selector_shift_output_projection_result,
            shape=[selector_shift_output_projection_result.shape[0], 1, 1],
        )
        # print("shift_output transitions reshaped: \n %s" % (selector_shift_output_projection_result_reshape.eval()))

        selector_shift_output_projection_result = (
            selector_shift_output_projection_result_reshape
            * selector_shift_output_filter
        )
        # print("shift_output transitions: \n %s" % selector_shift_output_projection_result.eval())
        return selector_shift_output_projection_result

    #############  ADD OUTPUT#######################

    def get_output_matrix(self, output_values):
        output_column = tf.reshape(
            tf.cast(output_values, dtype=tf.float32), shape=[self.batch_size, 1]
        )
        output_row = tf.constant(
            [
                [
                    0 if i != self.output_position[0] else 1
                    for i in range(0, self.state_size)
                ]
            ],
            dtype=tf.float32,
            shape=[1, self.state_size],
        )
        output_matrix = tf.matmul(
            output_column, output_row, a_is_sparse=True, b_is_sparse=True
        )
        output_matrix = tf.reshape(
            output_matrix, shape=[self.batch_size, self.state_size, 1]
        )
        return output_matrix

    ####################  ADD TRANSITION #######

    def get_transition_matrix(self, transitions_array):
        transition_column = tf.reshape(
            tf.cast(transitions_array, dtype=tf.float32), shape=[self.batch_size, 1]
        )
        transition_row = tf.constant(
            [
                [
                    0 if i != self.transition_position else 1
                    for i in range(0, self.state_size)
                ]
            ],
            dtype=tf.float32,
            shape=[1, self.state_size],
        )
        transition_matrix = tf.matmul(
            transition_column, transition_row, a_is_sparse=True, b_is_sparse=True
        )
        transition_matrix = tf.reshape(
            transition_matrix, shape=[self.batch_size, self.state_size, 1]
        )
        return transition_matrix

    ###################### MORPHEME ######################
    # TODO:esto parece ser solo la transicion shif_output!!

    def get_selector_morpheme_projection_result(
        self, transitions, transition_unit_vector
    ):
        selector_morpheme_projection = tf.constant(
            [transition_unit_vector],
            dtype=tf.float32,
            shape=[self.available_transitions_len, 1],
        )
        selector_morpheme_projection_result = tf.matmul(
            transitions, selector_morpheme_projection
        )
        return selector_morpheme_projection_result

    def get_selector_morpheme_filter(self, state_size):
        # 4
        selector_shift_output_filter = self.get_selector_shift_output_filter(
            state_size, self.output_position
        )
        selector_morpheme_filter = selector_shift_output_filter

        return selector_morpheme_filter

    def get_selector_morpheme_result(self, transitions, transition_unit_vector):

        selector_morpheme_filter = self.get_selector_morpheme_filter(self.state_size)
        selector_morpheme_filter = tf.reshape(
            tf.constant(selector_morpheme_filter, dtype=tf.float32),
            shape=[1, self.state_size, self.state_size],
        )
        # print("morpheme Filter: \n %s" % (selector_morpheme_filter.eval()))
        selector_morpheme_projection_result = self.get_selector_morpheme_projection_result(
            transitions, transition_unit_vector
        )
        selector_morpheme_projection_result_reshape = tf.reshape(
            selector_morpheme_projection_result,
            shape=[selector_morpheme_projection_result.shape[0], 1, 1],
        )
        # print("morpheme transitions reshaped: \n %s" % (selector_morpheme_projection_result_reshape.eval()))

        selector_morpheme_projection_result = (
            selector_morpheme_projection_result_reshape * selector_morpheme_filter
        )
        # print("morpheme transitions: \n %s" % selector_morpheme_projection_result.eval())
        return selector_morpheme_projection_result

    ###################### TRANSICIONES DISPONIBLES##################

    def get_one_hot_transitions(self):
        transitions_names_len = len(self.transition_names)
        available_transitions = np.zeros([transitions_names_len, transitions_names_len])
        for index, i in enumerate(available_transitions):
            available_transitions[index, index] = 1
        return available_transitions

    @staticmethod
    def get_transitions_array(next_transition):
        transitions_array = tf.argmax(next_transition, axis=1)
        return transitions_array

    @staticmethod
    def get_one_hot_index(one_hot):
        for index, i in enumerate(one_hot):
            if i == 1:
                return index


class TransitionSystemWorkerNode(WorkerNode):
    def __init__(
        self,
        stack_size: int,
        transition_size: int,
        output_size: int,
        input_size: int,
        batch_size: int,
        stack_position: int,
        stack_position_elems: list,
        transition_position: int,
        morpheme_position: list,
        input_position: int,
        execution_mode: ExecutionMode,
        transition_oracle: WorkerNode,
        output_oracle: WorkerNode,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        redis_connection_pool=None,
    ):
        WorkerNode.__init__(
            self,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )
        self.transition_system_configurations = TransitionSystemConfigurations(
            stack_size,
            transition_size,
            output_size,
            input_size,
            batch_size,
            stack_position,
            stack_position_elems,
            transition_position,
            morpheme_position,
            input_position,
            execution_mode,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )
        self.batch_size = batch_size
        self.transition_oracle = transition_oracle
        self.output_oracle = output_oracle
        # TODO:ajustar para que el parameternode me devuelva las cosas que necesito, asi el codgo es mas legible
        # self.filters, self.linear_regressor, self.options, self.temperature = self.transition_system_configurations()

    def _operation(
        self, *args
    ):  # recibe outputs array y state, creo que estos deberian ser parte de la clase, incluirlos como atributos
        if self.transition_system_configurations.execution_mode == ExecutionMode.eager:
            state_result = self.apply_eager_step(args[0][0])
        elif (
            self.transition_system_configurations.execution_mode == ExecutionMode.graph
        ):
            state_result = self.apply_graph_step(args[0][0])
            state_result = tf.expand_dims(state_result, axis=2)
        return state_result

    # aqui aplicamos solo la transicion del vector unitario que se recibe, por eso hay que iterar sobre las transiciones disponibles, esto no tiene mucho sentido ahora
    # quien llama a esta funcion debe ir acumulando el resultado
    def apply_transition(
        self,
        transitions,
        transition_unit_vector,
        state_size,
        original_state,
        current_state=None,
    ):
        transition_index = self.transition_system_configurations.get_one_hot_index(
            transition_unit_vector
        )
        original_state = tf.reshape(
            tf.cast(original_state, dtype=tf.float32),
            shape=[self.batch_size, state_size],
        )
        expanded_state = tf.expand_dims(original_state, axis=2)
        if transition_index == 0:  # shift
            # obtenemos el selctor con las matrices que se aplicaran
            selector_shift_result = self.transition_system_configurations.get_selector_shift_result(
                transitions, transition_unit_vector, state_size
            )
            new_state = tf.matmul(selector_shift_result, expanded_state)
        elif transition_index == 1:  # discar/clean_input
            selector_clean_input_projection_result = self.transition_system_configurations.get_selector_clean_input_result(
                transitions, transition_unit_vector, state_size
            )
            new_state = tf.matmul(
                selector_clean_input_projection_result, expanded_state
            )
        elif transition_index == 2:  # delete_stack
            selector_delete_stack_result = self.transition_system_configurations.get_selector_delete_stack_result(
                transitions, transition_unit_vector, state_size
            )
            new_state = tf.matmul(selector_delete_stack_result, expanded_state)
        elif transition_index == 3:  # shift_output
            selector_shift_output_result = self.transition_system_configurations.get_selector_shift_output_result(
                transitions, transition_unit_vector, state_size
            )
            new_state = tf.matmul(selector_shift_output_result, expanded_state)
        elif transition_index == 4:  # morpheme
            selector_morpheme_projection_result = self.transition_system_configurations.get_selector_morpheme_result(
                transitions, transition_unit_vector
            )
            new_state = tf.matmul(selector_morpheme_projection_result, expanded_state)
        else:  # esto es un hack para probar, falta inlcuir las otras transiciones en el if
            original_state_zeros = tf.constant(
                np.zeros(original_state.shape),
                dtype=tf.float32,
                shape=[self.batch_size, state_size],
            )
            new_state = tf.expand_dims(original_state_zeros, axis=2)

        if current_state is None:
            return new_state
        else:
            new_state = current_state + new_state
            return new_state

    # TODO: esta bien que esten hardcodes los unitarios en onehot correspondientes a cada transicion??
    # se calculan las transciones de un step a un batch, a diferencia de apply_transition no itera sino que recibe todos los vectores unitarios de transiciones disponibles
    def apply_batch_transition(
        self,
        transitions,
        transition_unit_vectors,
        state_size,
        original_state,
        current_state=None,
    ):
        original_state = (
            original_state  # priemra dimension es el numeor de rows en matriz de state
        )
        expanded_state = tf.expand_dims(original_state, axis=2)

        #######

        selector_shift_result = self.transition_system_configurations.get_selector_shift_result(
            transitions, list(transition_unit_vectors[0]), state_size
        )
        selector_clean_input_projection_result = self.transition_system_configurations.get_selector_clean_input_result(
            transitions, list(transition_unit_vectors[1]), state_size
        )
        selector_delete_stack_result = self.transition_system_configurations.get_selector_delete_stack_result(
            transitions, list(transition_unit_vectors[2]), state_size
        )
        selector_shift_output_result = self.transition_system_configurations.get_selector_shift_output_result(
            transitions, list(transition_unit_vectors[3]), state_size
        )
        selector_morpheme_projection_result = self.transition_system_configurations.get_selector_morpheme_result(
            transitions, list(transition_unit_vectors[4])
        )

        #######
        new_state = tf.matmul(selector_shift_result, expanded_state)
        new_state = new_state + tf.matmul(
            selector_clean_input_projection_result, expanded_state
        )
        new_state = new_state + tf.matmul(selector_delete_stack_result, expanded_state)
        new_state = new_state + tf.matmul(selector_shift_output_result, expanded_state)
        new_state = new_state + tf.matmul(
            selector_morpheme_projection_result, expanded_state
        )

        return new_state

    # next_transition: batch de transiciones
    def apply_eager_transitions(self, next_transition, original_state):
        state_size = self.transition_system_configurations.state_size
        state_result = None
        for transition in self.transition_system_configurations.available_transitions:
            self.transition_system_configurations.step_transition = next_transition
            if state_result is None:
                state_result = self.apply_transition(
                    next_transition, list(transition), state_size, original_state
                )
            else:
                transition_result = self.apply_transition(
                    next_transition,
                    list(transition),
                    state_size,
                    original_state,
                    current_state,
                )
                state_result = transition_result

            current_state = state_result
        self.transition_system_configurations.transitions_array = self.transition_system_configurations.get_transitions_array(
            next_transition
        )
        transition_matrix = self.transition_system_configurations.get_transition_matrix(
            self.transition_system_configurations.transitions_array
        )
        outputs_array = self.output_oracle(state_result, self.output_oracle.counter)
        output_matrix = self.transition_system_configurations.get_output_matrix(
            outputs_array
        )
        state_result = state_result + transition_matrix + output_matrix
        return state_result

    def apply_graph_transitions(self, next_transition, original_state, counter):
        state_size = self.transition_system_configurations.state_size
        state_result = None
        self.transition_system_configurations.step_transition = next_transition
        transition_unit_vectors = (
            self.transition_system_configurations.available_transitions
        )

        state_result = self.apply_batch_transition(
            next_transition, transition_unit_vectors, state_size, original_state
        )
        self.transition_system_configurations.transitions_array = self.transition_system_configurations.get_transitions_array(
            next_transition
        )
        transition_matrix = self.transition_system_configurations.get_transition_matrix(
            self.transition_system_configurations.transitions_array
        )
        outputs_array = self.output_oracle(state_result, counter)
        output_matrix = self.transition_system_configurations.get_output_matrix(
            outputs_array
        )
        state_result = state_result + transition_matrix + output_matrix
        return state_result

    def apply_eager_step(self, original_state):
        state = original_state
        condition = (
            self.transition_oracle.transitions.shape[0] > self.transition_oracle.counter
        )
        while condition:
            transitions = self.transition_oracle(state)  # transition o transitionsss??
            # transitions = tf.constant(transitions, dtype=tf.float32)
            self.transition_system_configurations.transitions = (
                transitions  # es un STEP, BATCH DE N ELEMENTOS
            )
            # output_array = self.output_oracle(state??)
            state = self.apply_eager_transitions(transitions, state)
            condition = (
                self.transition_oracle.transitions.shape[0]
                > self.transition_oracle.counter
            )
        return state

    # el counter es del oracle y lo comparo contra las transiciones, deberia ser otra la condicion??
    def transition_condition(self, counter, transitions_len, state):
        return tf.less(counter, transitions_len)

    def transition_body(self, counter, transitions_len, state):
        transitions = self.transition_oracle(
            state, counter
        )  # transition o transitionsss??
        self.transition_system_configurations.transitions = (
            transitions  # es un STEP, BATCH DE N ELEMENTOS
        )
        state = self.apply_graph_transitions(transitions, state, counter)
        state = tf.reshape(
            state,
            shape=[self.batch_size, self.transition_system_configurations.state_size],
        )
        counter = tf.add(counter, 1, name="current-counter")
        return [counter, transitions_len, state]

    def apply_graph_step(self, original_state):
        counter = tf.constant(
            self.transition_oracle.counter,
            dtype=tf.int32,
            shape=[],
            name="current-counter-init",
        )
        transitions_len = tf.constant(
            self.transition_oracle.transitions.shape[0],
            dtype=tf.int32,
            shape=[],
            name="transtions_len",
        )
        state = np.array(original_state, np.float32)
        counter, _, state = tf.while_loop(
            self.transition_condition,
            self.transition_body,
            [counter, transitions_len, state],
            parallel_iterations=1,
        )
        return state
